# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Contains information about each output stream."""

from streamer.bitrate_configuration import AudioCodec, AudioChannelLayout, VideoCodec, VideoResolution
from streamer.input_configuration import Input, MediaType
from streamer.pipe import Pipe
from typing import Dict, Union


class OutputStream(object):
  """Base class for output streams."""

  def __init__(self,
               type: MediaType,
               input: Input,
               codec: Union[AudioCodec, VideoCodec, None],
               pipe_dir: str,
               skip_transcoding: bool = False,
               pipe_suffix: str = '') -> None:

    self.type: MediaType = type
    self.skip_transcoding = skip_transcoding
    self.input: Input = input
    self.features: Dict[str, str] = {}
    self.codec: Union[AudioCodec, VideoCodec, None] = codec

    if self.skip_transcoding:
      # If skip_transcoding is specified, let the Packager read from a plain
      # file instead of an IPC pipe.
      self.ipc_pipe = Pipe.create_file_pipe(self.input.name, mode='r')
    else:
      self.ipc_pipe = Pipe.create_ipc_pipe(pipe_dir, pipe_suffix)

  def is_hardware_accelerated(self) -> bool:
    """Returns True if this output stream uses hardware acceleration."""
    if self.codec:
      return self.codec.is_hardware_accelerated()
    return False

  def get_ffmpeg_codec_string(self, hwaccel_api: str) -> str:
    """Returns a codec string accepted by FFmpeg for this stream's codec."""
    assert self.codec is not None
    return self.codec.get_ffmpeg_codec_string(hwaccel_api)

  def is_dash_only(self) -> bool:
    """Returns True if the output format is restricted to DASH protocol"""
    if self.codec is not None:
      return self.codec.get_output_format() == 'webm'
    return False

  def get_init_seg_file(self) -> Pipe:
    INIT_SEGMENT = {
      MediaType.AUDIO: 'audio_{language}_{channels}c_{bitrate}_{codec}_init.{format}',
      MediaType.VIDEO: 'video_{resolution_name}_{bitrate}_{codec}_init.{format}',
      MediaType.TEXT: 'text_{language}_init.{format}',
    }
    path_templ = INIT_SEGMENT[self.type].format(**self.features)
    return Pipe.create_file_pipe(path_templ, mode='w')

  def get_media_seg_file(self) -> Pipe:
    MEDIA_SEGMENT = {
      MediaType.AUDIO: 'audio_{language}_{channels}c_{bitrate}_{codec}_$Number$.{format}',
      MediaType.VIDEO: 'video_{resolution_name}_{bitrate}_{codec}_$Number$.{format}',
      MediaType.TEXT: 'text_{language}_$Number$.{format}',
    }
    path_templ = MEDIA_SEGMENT[self.type].format(**self.features)
    return Pipe.create_file_pipe(path_templ, mode='w')

  def get_single_seg_file(self) -> Pipe:
    SINGLE_SEGMENT = {
      MediaType.AUDIO: 'audio_{language}_{channels}c_{bitrate}_{codec}.{format}',
      MediaType.VIDEO: 'video_{resolution_name}_{bitrate}_{codec}.{format}',
      MediaType.TEXT: 'text_{language}.{format}',
    }
    path_templ = SINGLE_SEGMENT[self.type].format(**self.features)
    return Pipe.create_file_pipe(path_templ, mode='w')

  def get_identification(self) -> str:
    SINGLE_SEGMENT = {
      MediaType.AUDIO: '{language}_{channels}c_{bitrate}_{codec}_{format}',
      MediaType.VIDEO: '{resolution_name}_{bitrate}_{codec}_{format}',
      MediaType.TEXT: '{language}_{format}',
    }
    return SINGLE_SEGMENT[self.type].format(**self.features)


class AudioOutputStream(OutputStream):

  def __init__(self,
               input: Input,
               pipe_dir: str,
               codec: AudioCodec,
               channel_layout: AudioChannelLayout) -> None:

    super().__init__(MediaType.AUDIO, input, codec, pipe_dir)
    # Override the codec type and specify that it's an audio codec
    self.codec: AudioCodec = codec
    self.layout = channel_layout

    # The features that will be used to generate the output filename.
    self.features = {
      'language': input.language,
      'channels': str(self.layout.max_channels),
      'bitrate': self.get_bitrate(),
      'format': self.codec.get_output_format(),
      'codec': self.codec.value,
    }

  def get_bitrate(self) -> str:
    """Returns the bitrate for this stream."""
    return self.layout.bitrates[self.codec]


class VideoOutputStream(OutputStream):

  def __init__(self,
               input: Input,
               pipe_dir: str,
               codec: VideoCodec,
               resolution: VideoResolution) -> None:
    super().__init__(MediaType.VIDEO, input, codec, pipe_dir)
    # Override the codec type and specify that it's an audio codec
    self.codec: VideoCodec = codec
    self.resolution = resolution

    # The features that will be used to generate the output filename.
    self.features = {
      'resolution_name': self.resolution.get_key(),
      'bitrate': self.get_bitrate(),
      'format': self.codec.get_output_format(),
      'codec': self.codec.value,
    }

  def get_bitrate(self) -> str:
    """Returns the bitrate for this stream."""
    return self.resolution.bitrates[self.codec]


class TextOutputStream(OutputStream):

  def __init__(self,
               input: Input,
               pipe_dir: str,
               skip_transcoding: bool):
    # We don't have a codec per se for text, but we'd like to generically
    # process OutputStream objects in ways that are easier with this attribute
    # set, so set it to None.
    codec = None

    super().__init__(MediaType.TEXT, input, codec, pipe_dir,
                     skip_transcoding, pipe_suffix='.vtt')

    # The features that will be used to generate the output filename.
    self.features = {
      'language': input.language,
      'format': 'mp4',
    }



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A simple proxy server to upload to cloud stroage providers."""

import time
import traceback
import urllib.parse

from http.server import ThreadingHTTPServer, BaseHTTPRequestHandler
from typing import Optional, Union

from streamer.node_base import ProcessStatus, ThreadedNodeBase

from streamer.cloud.pool import Pool
from streamer.cloud.uploader import ALL_SUPPORTED_PROTOCOLS, SUPPORTED_PROTOCOLS


# HTTP status codes
HTTP_STATUS_CREATED = 201
HTTP_STATUS_ACCEPTED = 202
HTTP_STATUS_NO_CONTENT = 204
HTTP_STATUS_FAILED = 500


# Don't write the same file more than once per rate limiter period.
# For live streams, this avoids HTTP 429 "Too many request" errors.
RATE_LIMITER_PERIOD_IN_SECONDS = 2


class RateLimiter(object):
  """A rate limiter that tracks which files we have written to recently."""

  def __init__(self) -> None:
    self._reset(time.time())

  def suppress(self, path) -> bool:
    """Returns true if you should skip this upload."""

    now = time.time()
    if now > self._last_check + RATE_LIMITER_PERIOD_IN_SECONDS:
      self._reset(now)

    if path in self._recent_files:
      return True  # skip

    self._recent_files.add(path)
    return False  # upload

  def _reset(self, now: float) -> None:
    # These files are only valid for RATE_LIMITER_PERIOD_IN_SECONDS.
    # After that, they get cleared.
    self._recent_files: set[str] = set()

    # The timestamp of the last check; the start of the rate limiter period.
    self._last_check: float = now


class RequestHandler(BaseHTTPRequestHandler):
  """A request handler that processes requests coming from Shaka Packager and
  relays them to the destination.
  """

  def __init__(self, rate_limiter: RateLimiter, pool: Pool,
               *args, **kwargs) -> None:
    self._rate_limiter: RateLimiter = rate_limiter
    self._pool: Pool = pool

    # The HTTP server passes *args and *kwargs that we need to pass along, but
    # don't otherwise care about.  This must happen last, or somehow our
    # members never get set.
    super().__init__(*args, **kwargs)

  # NOTE: The default values here for log_request are taken from the base
  # class, and not a design decision of ours.
  def log_request(self, code: Union[int, str] = '-', size: Union[int, str] = '-') -> None:
    """Override the request logging feature of the Python HTTP server."""
    try:
      code_int = int(code)
    except:
      code_int = 0

    if code_int >= 200 and code_int <= 299:
      # Stub out log_request to avoid creating noise from the HTTP server when
      # requests are successful.
      return

    return super().log_request(code, size)

  def _parse_chunked_transfer(self, suppress: bool) -> None:
    # Here we parse the chunked transfer encoding and delegate to the
    # subclass's start/chunk/end methods.  If |suppress|, we parse the input
    # but don't do anything with it.
    with self._pool.get_worker() as worker:
      if not suppress:
        worker.start_chunked(self.path)

      while True:
        # Parse the chunk size
        chunk_size_line = self.rfile.readline().strip()
        chunk_size = int(chunk_size_line, 16)

        # Read the chunk and process it
        if chunk_size != 0:
          data = self.rfile.read(chunk_size)
          if not suppress:
            worker.write_chunk(data)

        self.rfile.readline()  # Read the trailer

        if chunk_size == 0:
           break  # EOF

      # All done.
      if not suppress:
        worker.end_chunked()

  def _parse_non_chunked_transfer(self, suppress: bool) -> None:
    # We have the whole file at once, with a known length.
    content_length = int(self.headers['Content-Length'])

    if suppress:
      # If |suppress|, we read the input but don't do anything with it.
      self.rfile.read(content_length)
    else:
      with self._pool.get_worker() as worker:
        worker.write_non_chunked(self.path, self.rfile.read(content_length))

  def do_PUT(self) -> None:
    """Handle PUT requests coming from Shaka Packager."""
    suppress = self._rate_limiter.suppress(self.path)

    try:
      if self.headers.get('Transfer-Encoding', '').lower() == 'chunked':
        self._parse_chunked_transfer(suppress)
      else:
        self._parse_non_chunked_transfer(suppress)

      # Close the input and respond.
      self.rfile.close()
      self.send_response(HTTP_STATUS_ACCEPTED if suppress else HTTP_STATUS_CREATED)
    except Exception as ex:
      print('Upload failure: ' + str(ex))
      traceback.print_exc()
      self.send_response(HTTP_STATUS_FAILED)

    # If we don't call this at the end of the handler, Packager says we
    # "returned nothing".
    self.end_headers()

  def do_DELETE(self) -> None:
    """Handle DELETE requests coming from Shaka Packager."""
    try:
      with self._pool.get_worker() as worker:
        worker.delete(self.path)
      self.send_response(HTTP_STATUS_NO_CONTENT)
    except Exception as ex:
      print('Upload failure: ' + str(ex))
      traceback.print_exc()
      self.send_response(HTTP_STATUS_FAILED)

    # If we don't call this at the end of the handler, Packager says we
    # "returned nothing".
    self.end_headers()


class ProxyNode(ThreadedNodeBase):
  """Runs an HTTP server at `self.server_location` to upload to cloud.

  Subclasses handle upload to specific cloud storage providers.

  The local HTTP server at `self.server_location` can only ingest PUT requests.
  """
  SUPPORTED_PROTOCOLS = SUPPORTED_PROTOCOLS
  ALL_SUPPORTED_PROTOCOLS = ALL_SUPPORTED_PROTOCOLS

  server_location: str = ''

  def __init__(self, upload_location: str, pool_size: int) -> None:
    super().__init__(thread_name=self.__class__.__name__,
                     continue_on_exception=True,
                     sleep_time=3)
    if not ProxyNode.is_supported(upload_location):
      raise RuntimeError("Protocol of {} isn't supported".format(upload_location))

    self._upload_location = upload_location
    self._rate_limiter = RateLimiter()
    self._server: Optional[ThreadingHTTPServer] = None
    self._pool: Optional[Pool] = None
    self._pool_size: int = pool_size

  def create_handler(self, *args, **kwargs) -> BaseHTTPRequestHandler:
    assert self._pool is not None
    return RequestHandler(self._rate_limiter, self._pool, *args, **kwargs)

  def start(self) -> None:
    # Will be started early to get server location.
    if self._server is not None:
      return

    self._pool = Pool(self._upload_location, self._pool_size)

    handler_factory = (
        lambda *args, **kwargs: self.create_handler(*args, **kwargs))

    # By specifying port 0, a random unused port will be chosen for the server.
    self._server = ThreadingHTTPServer(
        ('localhost', 0), handler_factory)
    self.server_location = (
        'http://' + self._server.server_name +
        ':' + str(self._server.server_port))

    return super().start()

  def stop(self, status: Optional[ProcessStatus]) -> None:
    if self._server:
      self._server.shutdown()
      self._server = None
    if self._pool:
      self._pool.close()
      self._pool = None
    return super().stop(status)

  def check_status(self) -> ProcessStatus:
    # This makes sure this node will never prevent the shutdown of the whole
    # system.  It will be stopped explicitly when ControllerNode tears down.
    return ProcessStatus.Finished

  def _thread_single_pass(self) -> None:
    assert self._server is not None
    # Will terminate on server.shutdown().
    self._server.serve_forever()

  @staticmethod
  def is_understood(upload_location: str) -> bool:
    """Is the URL understood, independent of libraries available?"""
    url = urllib.parse.urlparse(upload_location)
    return url.scheme in ALL_SUPPORTED_PROTOCOLS

  @staticmethod
  def is_supported(upload_location: str) -> bool:
    """Is the URL supported with the libraries available?"""
    url = urllib.parse.urlparse(upload_location)
    return url.scheme in SUPPORTED_PROTOCOLS



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Contains the helper classes for HLS parsing and concatenation."""

import os
import re
import math
import posixpath
from typing import List, Dict, Set, Optional, Tuple
from streamer.output_stream import AudioOutputStream, OutputStream, TextOutputStream, VideoOutputStream
from streamer.bitrate_configuration import VideoCodec, AudioCodec, VideoResolution, AudioChannelLayout
from streamer.packager_node import PackagerNode


class MediaPlaylist:
  """A class representing a media playlist(any playlist that references
  media files or media files' segments whether they were video/audio/text files).
  
  The information collected from the master playlist about
  a specific media playlist such as the bitrate, codec, etc...
  is also stored here in `self.stream_info` dictionary.
  
  Keep in mind that a stream variant playlist is also a MediaPlaylist.
  """
  
  current_stream_index = 0
  """A number that is shared between all the MediaPlaylist objects to be used
  to generate unique file names in the format 'stream_<current_stream_index>.m3u8'
  """
  
  HEADER_TAGS = ('#EXTM3U', '#EXT-X-VERSION', '#EXT-X-PLAYLIST-TYPE')
  """Common header tags to search for so we don't store them
  in `MediaPlaylist.content`.  These tags must be defined only one time
  in a playlist, and written at the top of a media playlist file once.
  """
  
  def __init__(self,
               stream_info: Dict[str, str],
               dir_name: Optional[str] = None,
               output_dir: Optional[str] = None,
               streams_map: Optional[Dict[str, OutputStream]] = None):
    """Given a `stream_info` and the `dir_name`, this method finds the media
    playlist file, parses it, and stores relevant parts of the playlist in
    `self.content`.
    
    It also updates the segment paths to make it relative to `output_dir`
    
    A `streams_map` is used to match this media playlist to its OutputStream
    object.
    """
    
    self.stream_info = stream_info
    self.duration = 0.0
    self.target_duration = 0
    
    self.content = ''
    
    if dir_name is None:
      # Do not read, The content will be added manually.
      return
    
    # If there is a file to read, we MUST have a streams_map to match this
    # media playlist file with its OutputStream.
    assert streams_map is not None
    
    period_dir = os.path.relpath(dir_name, output_dir)
    media_playlist_file = os.path.join(dir_name,
                                       _unquote(self.stream_info['URI']))
    
    with open(media_playlist_file) as media_playlist:
      line = media_playlist.readline()
      while line:
        if line.startswith('#EXTINF'):
          # Add this segment duration to the total duration.
          # This will be used to re-calculate the average bitrate.
          self.duration += float(line[len('#EXTINF:'):].split(',', 1)[0])
          self.content += line
          line = media_playlist.readline()
          # If a byterange exists, add it to the content.
          if line.startswith('#EXT-X-BYTERANGE'):
            self.content += line
            line = media_playlist.readline()
          # Update the segment's URI.
          self.content += posixpath.join(period_dir ,line)
        elif line.startswith('#EXT-X-MAP'):
          # An EXT-X-MAP must have a URI attribute and optionally
          # a BYTERANGE attribute.
          attribs = _extract_attributes(line)
          self.content += '#EXT-X-MAP:URI=' + _quote(
              posixpath.join(period_dir, _unquote(attribs['URI'])))
          if attribs.get('BYTERANGE'):
            self.content += ',BYTERANGE=' + attribs['BYTERANGE']
          self.content += '\n'
        elif line.startswith(MediaPlaylist.HEADER_TAGS + ('#EXT-X-ENDLIST',)):
          # Skip header and end-list tags.
          pass
        elif not line.startswith('#EXT'):
          # Skip comments.
          pass
        elif line.startswith('#EXT-X-TARGETDURATION'):
          self.target_duration = int(line[len('#EXT-X-TARGETDURATION:'):])
        else:
          # Store lines that didn't match one of the above cases.
          # Like ENCRYPTIONKEYS, DISCONTINUITIES, COMMENTS, etc... .
          self.content += line
        line = media_playlist.readline()
    
    # Set the features we need to access easily while performing the concatenation.
    # Features like codec, channel_layout, resolution, etc... .
    self._set_features(streams_map)
  
  def _set_features(self, streams_map: Dict[str, OutputStream]) -> None:
    """Get the audio and video codecs and other relevant stream features
    from the matching OutputStream in the `streams_map`, this will be used
    in the codec matching process in the concat_xxx() methods, but the codecs
    that will be written in the final concatenated master playlist will be
    the codecs from stream_info dictionary(in HLS syntax).
    """
    
    # We can't depend on the stream_info['CODECS'] to get the codecs from because
    # this is only present for STREAM-INF, this makes it harder to get codecs for
    # audio segments. Also if we try to get the audio codecs from one of the 
    # #EXT-X-STREAM-INF tags we will have to match these codecs with each stream,
    # for example, a codec attribute might be CODECS="avc1,ac-3,mp4a,opus",
    # the video codec is put first then the audio codecs are put in 
    # lexicographical order(by observation), which isn't necessary the same order of 
    # #EXT-X-MEDIA in the master playlist, thus there is no solid baseground for 
    # matching the codecs using the information in the master playlist.
    
    output_stream: Optional[OutputStream] = None
    lines = self.content.split('\n')
    # We need to seek to the first line with the tag #EXTINF.  The line after
    # will have the URI we need, but if we encounter a byterange tag we need to
    # advance one more line.
    for i in range(len(lines)):
      # Don't use the URIs from any tag to try to extract codec information.
      # We should not rely on the exact structure of file names for this.
      # Use stream_maps instead.
      line = lines[i]
      if line.startswith('#EXTINF'):
        line = lines[i + 1]
        if line.startswith('#EXT-X-BYTERANGE'):
          line = lines[i + 2]
        file_name = os.path.basename(line)
        # Index the file_name and don't use dict.get() .
        # There MUST be a match.
        output_stream = streams_map[file_name]
        break
    
    assert output_stream, 'No media file found in this media playlist'
    self.codec = output_stream.codec
    if isinstance(output_stream, VideoOutputStream):
      self.resolution = output_stream.resolution
    elif isinstance(output_stream, AudioOutputStream):
      self.channel_layout = output_stream.layout
      # We will get the language from the stream_info because the stream information
      # is provided by Packager.  We might have mixed 3-letter and 2-letter format
      # from the Streamer, but Packager reduces them all to 2-letter language tag.
      self.language = _unquote(self.stream_info.get('LANGUAGE', '"und"'))
    elif isinstance(output_stream, TextOutputStream):
      self.language = _unquote(self.stream_info.get('LANGUAGE', '"und"'))
  
  def write(self, dir_name: str, media_playlist_header: str):
    """Writes a media playlist whose file name is `self.stream_info['URI']`
    in the directory `dir_name`.
    """
    
    file_path = os.path.join(dir_name, _unquote(self.stream_info['URI']))
    with open(file_path, 'w') as media_playlist_file:
      content = media_playlist_header
      content += '#EXT-X-TARGETDURATION:' + str(self.target_duration) + '\n\n'
      content += self.content
      content += '#EXT-X-ENDLIST\n'
      media_playlist_file.write(content)
  
  @staticmethod
  def extract_header(file_path: str) -> str:
    """Extracts the common media playlist header(parts we can't store
    in `self.content`).  We then write this header once at the top of
    the file when MediaPlaylist.write() is called.
    """
    
    header = ''
    with open(file_path, 'r') as media_playlist:
      line = media_playlist.readline()
      while line:
        # Capture the M3U tag, PlaylistType, and ExtVersion.
        if line.startswith(MediaPlaylist.HEADER_TAGS):
          header += line
        line = media_playlist.readline()
    return header
  
  @staticmethod
  def _similar_stream_info(media_playlists: List['MediaPlaylist']
                                ) -> Dict[str, str]:
    """A helper method, used to return the stream_info values that are
    identical across all the streams.
    
    A non-empty list must be passed for the parameter `media_playlists`.
    """
    
    assert len(media_playlists), ('There MUST be at least one media playlist'
                                  'to collect its stream information')
    # Get an arbitrary stream info.
    stream_info = media_playlists[0].stream_info.copy()
    
    for key in list(stream_info):
      for media_playlist in media_playlists:
        # If a media playlist has a different value for this key, pop it.
        if media_playlist.stream_info.get(key) != stream_info[key]:
          stream_info.pop(key)
          break
    
    return stream_info
  
  @staticmethod
  def _get_bandwidth(var_playlists: List['MediaPlaylist'],
                     durations: List[float]) -> Dict[str, str]:
    """A helper method to get the peak and average bandwidth
    for stream variants.
    
    For the BANDWIDTH we pick the maximum one we have.
    BANDWIDTH = max((BANDWIDTH)s)
    
    For the AVERAGE-BANDWIDTH we perform a weighted average by duration.
    AVERAGE-BANDWIDTH = (AVERAGE-BANDWIDTH)s . (DURATION)s
                        /sum((DURATION)s)
    """
    
    band, avg_band = 0, 0.0
    for var_playlist, duration in zip(var_playlists, durations):
      band = max(int(var_playlist.stream_info['BANDWIDTH']), band)
      avg_band += int(var_playlist.stream_info['AVERAGE-BANDWIDTH']) * duration
    
    return {
        'BANDWIDTH': str(band),
        'AVERAGE-BANDWIDTH': str(math.ceil(avg_band/sum(durations)))
      }
  
  @staticmethod
  def _get_hls_codec(var_playlists: List['MediaPlaylist']) -> str:
    """A helper method to get all the possible codecs for a concatenated stream
    variant, which is all the codecs present in all the variants(in different
    periods) that will be concatenated.
    """
    
    codec_strings: Set[str] = set()
    for var_playlist in var_playlists:
      codec_strings.update(_unquote(
          var_playlist.stream_info['CODECS']).split(','))
    
    return _quote(','.join(codec_string for
                           codec_string in codec_strings))
  
  @staticmethod
  def _next_unique_name() -> Dict[str, str]:
    """Returns a unique NAME and URI."""
    
    stream_name = 'stream_' + str(MediaPlaylist.current_stream_index)
    MediaPlaylist.current_stream_index += 1
    
    return {
        'NAME': _quote(stream_name),
        'URI': _quote(stream_name + '.m3u8')
      }
  
  @staticmethod
  def _max_channels(aud_playlists: List['MediaPlaylist']
                    ) -> Dict[str, str]:
    """Returns the maximum channel count in the given audio playlists."""
    
    return {
        'CHANNELS': _quote(str(max(
            aud_playlist.channel_layout.max_channels for
            aud_playlist in aud_playlists
      )))}
  
  @staticmethod
  def _max_target_dur(media_playlists: List['MediaPlaylist']) -> int:
    """Returns the maximum target duration in the given media playlists."""
    
    return max(media_playlist.target_duration for
               media_playlist in media_playlists)
  
  @staticmethod
  def _fit_missing_lang(variant_options: List['MediaPlaylist'],
                        language: str) -> str:
    """Returns a substitution language for a missing language by considering
    the languages of the given variants.
    
    `variant_options` are audio or text playlists for the same input that
    we can choose a language from.
    
    `language` is the language this method will try to find a best fit for.
    """
    
    # Set the best_fit to be an arbitrary language for now.
    best_fit = variant_options[0].language
    language_base = language.split('-', 1)[0]
    
    for variant in variant_options:
      best_fit_base = best_fit.split('-', 1)[0]
      candidate = variant.language
      candidate_split = candidate.split('-', 1)
      candidate_base = candidate_split[0]
      candidate_is_regional = len(candidate_split) > 1
      # Only kick the previous best fit out when: The base of the candidate 
      # is the same as the base of the original language AND (the base of 
      # the best fit is not the same as the base of the original language
      # OR the candidate is a regional variant).
      if language_base == candidate_base:
        if language_base != best_fit_base or candidate_is_regional:
          best_fit = candidate
      # Note that no perfect match would ever occur, as this method 
      # is called only when a perfect match is missing.
    
    return best_fit
  
  @staticmethod
  def concat_sub(all_txt_playlists: List[List['MediaPlaylist']],
                 durations: List[float]) -> List['MediaPlaylist']:
    """A method that concatenates subtitle streams based on the language of the subtitles
    for each period, it will concatenate all the english subtitles, french subtitles, etc...
    and put them ordered in periods where a discontinuity is inserted between them.
    
    A `x=List['MediaPlaylists']` is returned, where `len(x)` is the total 
    number of langauges found, and the number of periods(discontinuities) in each playlist
    is `len(all_txt_playlists)`.
    
    When no subtitles are found for a specific langauge for a specific period, we
    try to substitute for it with another language, when no substitution is possible, 
    we insert an empty WEBVTT content as a filler for this period's duration.
    
    All the language un-annotated streams for each period gets concatenated together.
    """
    
    def non_nones(items: List[Optional[MediaPlaylist]]) -> List[MediaPlaylist]:
      """Return the elements of the list which are not None."""
      
      return [item for item in items if item is not None]
    
    # Extract all the languages from every stream.
    langs: Set[str] = set()
    for txt_playlists in all_txt_playlists:
      for txt_playlist in txt_playlists:
        langs.add(txt_playlist.language)
    
    # Create and initialize a division map to divide streams of different
    # languages into.
    division: Dict[str, List[Optional[MediaPlaylist]]] = {}
    for lang in langs:
      division[lang] = []
      # Assume all media playlists for lang are not there at the begining.
      for _ in range(len(all_txt_playlists)):
        # None = no media playlist with this language for this period.
        division[lang].append(None)
    
    # Add the ith media playlist in the ith position for its language division.
    for i, txt_playlists in enumerate(all_txt_playlists):
      for txt_playlist in txt_playlists:
        # Assume that for any langauge, there is at most one stream for it.
        # NOTE: A duplicate second language will overwrite the first one.
        division[txt_playlist.language][i] = txt_playlist
    
    # Try to substitute for missing languages if possible.
    for i in range(len(all_txt_playlists)):
      # All the language options that we can use to substitute for a missing
      # language in this period.
      txt_playlist_options = non_nones([division[lg][i] for lg in langs])
      # We can query `_fit_missing_lang()` only if we have some text
      # streams available, otherwise, we leave them Nones as they are.
      if len(txt_playlist_options):
        for lang in langs:
          if not division[lang][i]:
              sub_lang = MediaPlaylist._fit_missing_lang(txt_playlist_options,
                                                         lang)
              division[lang][i] = division[sub_lang][i]
    
    concat_txt_playlists: List[MediaPlaylist] = []
    for lang, optional_txt_playlists in division.items():
      stream_info = MediaPlaylist._similar_stream_info(
          # There must be at least one that isn't None.
          non_nones(optional_txt_playlists))
      if lang != 'und':
        # Put the language attribute in case it was removed.
        stream_info['LANGUAGE'] = _quote(lang)
      # Get a unique NAME and URI.
      stream_info.update(MediaPlaylist._next_unique_name())
      concat_txt_playlist = MediaPlaylist(stream_info)
      # Set the target duration of the concat playlist to the max of 
      # all children playlists.
      concat_txt_playlist.target_duration = MediaPlaylist._max_target_dur(
          non_nones(optional_txt_playlists))
      for i, optional_txt_playlist in enumerate(optional_txt_playlists):
        if optional_txt_playlist:
          # If a playlist is there, append it.
          concat_txt_playlist.content += optional_txt_playlist.content
        else:
          # If no playlist were found for this period, we create a time gap
          # by filling the period's duration with an empty string.
          ext_inf_count = math.ceil(durations[i] /
                                    concat_txt_playlist.target_duration)
          for _ in range(ext_inf_count):
            concat_txt_playlist.content += (
                '#EXTINF:' + str(durations[i] / ext_inf_count) + ',\n' +
                'data:text/vtt;charset=utf-8,WEBVTT%0A%0A\n')
        # Add a discontinuity after each period.
        concat_txt_playlist.content += '#EXT-X-DISCONTINUITY\n\n'
      concat_txt_playlists.append(concat_txt_playlist)
    
    return concat_txt_playlists
  
  @staticmethod
  def concat_aud_common(all_aud_playlists: List[List['MediaPlaylist']]
                        ) -> Dict[AudioCodec,
                                  Dict[str,
                                       Dict[AudioChannelLayout,
                                            List['MediaPlaylist']]]]:
    """A common method that is used to divide the audio playlists into a structure
    that is easy to process.
    
    This common method is used by `concat_aud` and `concat_aud_only` methods.
    """
    
    # Extract all the codecs, languages, and channel layouts.
    codecs: Set[AudioCodec] = set()
    langs: Set[str] = set()
    channels: Set[AudioChannelLayout] = set()
    for aud_playlists in all_aud_playlists:
      for aud_playlist in aud_playlists:
        assert isinstance(aud_playlist.codec, AudioCodec)
        codecs.add(aud_playlist.codec)
        langs.add(aud_playlist.language)
        channels.add(aud_playlist.channel_layout)
    
    # Create and initialize a division map.
    division: Dict[AudioCodec,
                   Dict[str,
                        Dict[AudioChannelLayout,
                             List['MediaPlaylist']]]] = {}
    
    for codec in codecs:
      division[codec] = {}
      for lang in langs:
        division[codec][lang] = {}
        for channel in channels:
          division[codec][lang][channel] = []
    
    # This logic inside here is done on period basis.
    for aud_playlists in all_aud_playlists:
      # Initialize a mapping between audio codecs and language to a list of
      # media playlists with channel layouts available.
      codec_lang_division: Dict[AudioCodec, Dict[str, List[MediaPlaylist]]] = {}
      for codec in codecs:
        codec_lang_division[codec] = {}
        for lang in langs:
          codec_lang_division[codec][lang] = []
      # For every audio playlist in this period, append it to the matching 
      # codec/language.
      for aud_playlist in aud_playlists:
        assert isinstance(aud_playlist.codec, AudioCodec)
        codec_lang_division[aud_playlist.codec][aud_playlist.language].append(
            aud_playlist)
      # Sort and replace the missing languages in the codec_lang_division map.
      for codec in codecs:
        for lang in langs:
          # If this language for this codec in this period has no media playlists
          # for any channel layout, this means that the language itself
          # is missing.  We will try to find a substitution for it.
          if not len(codec_lang_division[codec][lang]):
            aud_playlist_options = [codec_lang_division[codec][lang][0] for
                                    lang in langs
                                    if len(codec_lang_division[codec][lang])]
            sub_lang = MediaPlaylist._fit_missing_lang(aud_playlist_options,
                                                       lang)
            # Replace the empty codec_lang_division[codec][lang] with the
            # substitution language with the same codec.
            codec_lang_division[codec][lang] = codec_lang_division[codec][sub_lang]
          # Sort the media playlists ascendingly based on the channel layouts.
          codec_lang_division[codec][lang].sort(key=lambda pl: pl.channel_layout)
          # Fill the division map for the current period from the codec_lang_division map.
          for i, channel in enumerate(sorted(channels)):
            division[codec][lang][channel].append(
                # We will try to append the ith audio playlist which has
                # channel layout of `channel`(the for loop variable), but
                # if we don't have the ith audio playlist, we can substitute
                # for it with the max audio playlist in terms of channel layout.
                # This would be an optimal substitution since it is performed only
                # at the higher channel layouts, while the lower channel layouts
                # MUST always align since they had a shared pipeline configuration.
                codec_lang_division[codec][lang][min(
                    i, len(codec_lang_division[codec][lang]) - 1)])
    
    return division
  
  @staticmethod
  def concat_aud(all_aud_playlists: List[List['MediaPlaylist']]
                 ) -> List['MediaPlaylist']:
    """Concatenates multiple audio playlists into one multi-period playlist for
    different languages, codecs, and channel layouts.
    """
    
    division = MediaPlaylist.concat_aud_common(all_aud_playlists)
    
    concat_aud_playlists: List[MediaPlaylist] = []
    for codec, lang_channel_div in division.items():
      for lang, channel_div in lang_channel_div.items():
        for channel, aud_playlists in channel_div.items():
          stream_info = MediaPlaylist._similar_stream_info(
              aud_playlists)
          if lang != 'und':
            # Put the language attribute in case it was removed.
            stream_info['LANGUAGE'] = _quote(lang)
          # Get a unique file name.
          stream_info.update(MediaPlaylist._next_unique_name())
          # Set the max channels for this playlist.
          stream_info.update(MediaPlaylist._max_channels(aud_playlists))
          concat_aud_playlist = MediaPlaylist(stream_info)
          # Set the target duration.
          concat_aud_playlist.target_duration = MediaPlaylist._max_target_dur(
              aud_playlists)
          for aud_playlist in aud_playlists:
            concat_aud_playlist.content += aud_playlist.content
            # Add a discontinuity after each period.
            concat_aud_playlist.content += '#EXT-X-DISCONTINUITY\n\n'
          concat_aud_playlists.append(concat_aud_playlist)
    
    return concat_aud_playlists
  
  @staticmethod
  def concat_aud_only(all_aud_playlists: List[List['MediaPlaylist']],
                      all_var_playlists: List[List['MediaPlaylist']],
                      durations: List[float]) -> List['MediaPlaylist']:
    """Concatenates audio only periods with other audio only periods."""
    
    # Pair audio streams with their equivalent stream variants
    # to retrieve them back later.
    pair: Dict[MediaPlaylist, MediaPlaylist] = {}
    for aud_playlists, var_playlists in zip(all_aud_playlists,
                                            all_var_playlists):
      for aud_playlist in aud_playlists:
        # Search for the matching stream variant.
        for var_playlist in var_playlists:
          if var_playlist.stream_info['URI'] == aud_playlist.stream_info['URI']:
            pair[aud_playlist] = var_playlist
            break
    
    division = MediaPlaylist.concat_aud_common(all_aud_playlists)
    
    concat_aud_only_playlists: List[MediaPlaylist] = []
    for codec, lang_channel_div in division.items():
      for lang, channel_div in lang_channel_div.items():
        for channel, aud_playlists in channel_div.items():
          # Get the variant playlists paired to the current audio playlists.
          var_playlists = [pair[aud_playlist] for aud_playlist in aud_playlists]
          stream_info = MediaPlaylist._similar_stream_info(
              aud_playlists)
          if lang != 'und':
            # Put the language attribute in case it was removed.
            stream_info['LANGUAGE'] = _quote(lang)
          # Get a unique file name.
          stream_info.update(MediaPlaylist._next_unique_name())
          # Set the max channels for this playlist.
          stream_info.update(MediaPlaylist._max_channels(aud_playlists))
          concat_aud_playlist = MediaPlaylist(stream_info)
          # Get the stream info for the paired variant playlists.
          stream_info = MediaPlaylist._similar_stream_info(var_playlists)
          # The URI for the concatenated variant playlist will be the same as
          # the concatenated audio playlist.
          stream_info['URI'] = concat_aud_playlist.stream_info['URI']
          # Get the peak and average bandwidth across all periods
          # for this codec-language-channel triad.
          stream_info.update(MediaPlaylist._get_bandwidth(var_playlists,
                                                          durations))
          # Get the codecs from the associated stream variant playlists.
          stream_info['CODECS'] = MediaPlaylist._get_hls_codec(var_playlists)
          concat_var_playlist = MediaPlaylist(stream_info)
          # Set the target duration.
          concat_aud_playlist.target_duration = MediaPlaylist._max_target_dur(
              aud_playlists)
          for aud_playlist in aud_playlists:
            concat_aud_playlist.content += aud_playlist.content
            # Add a discontinuity after each period.
            concat_aud_playlist.content += '#EXT-X-DISCONTINUITY\n\n'
          # The audio and the stream variant playlist will be exactly the same.
          concat_var_playlist.target_duration = concat_aud_playlist.target_duration
          concat_var_playlist.content = concat_aud_playlist.content
          concat_aud_only_playlists.extend(
              [concat_aud_playlist, concat_var_playlist])
    
    return concat_aud_only_playlists
  
  @staticmethod
  def concat_vid(all_vid_playlists: List[List['MediaPlaylist']],
                 durations: List[float]) -> List['MediaPlaylist']:
    """Concatenates multiple video playlists into one multi-period playlist
    for many resolutions and codecs, it matches the codecs first, then matches
    the resolutions.
    
    It will pick the closest lower resolution for some period(input) if a high
    enough resolution was not available.
    
    A video playlist is a stream variant playlist based on the Packager's output.
    """
    
    # Get all possible video codecs.  We should not use the all codecs available
    # in the pipeline config, because for some codecs we might have given the
    # Packager the 'dash_only' flag.
    codecs: Set[VideoCodec] = set()
    # Also get all the available resolutions.
    resolutions: Set[VideoResolution] = set()
    for vid_playlists in all_vid_playlists:
      for vid_playlist in vid_playlists:
        assert isinstance(vid_playlist.codec, VideoCodec)
        codecs.add(vid_playlist.codec)
        resolutions.add(vid_playlist.resolution)
    
    # Create and initialize a division map.
    division: Dict[VideoCodec,
                   Dict[VideoResolution,
                        List[MediaPlaylist]]] = {}
    for codec in codecs:
      division[codec] = {}
      for resolution in resolutions:
        division[codec][resolution] = []
    
    # In each period do the following:
    for vid_playlists in all_vid_playlists:
      # Initialize a mapping between video codecs and a list of resolutions available.
      codec_division: Dict[VideoCodec, List[MediaPlaylist]] = {}
      for codec in codecs:
        codec_division[codec] = []
      # For every video playlist in this period, append it to the matching codec.
      for vid_playlist in vid_playlists:
        assert isinstance(vid_playlist.codec, VideoCodec)
        codec_division[vid_playlist.codec].append(vid_playlist)
      for codec in codecs:
        # Sort the variants from low resolution to high resolution.
        codec_division[codec].sort(key=lambda pl: pl.resolution)
        for i, resolution in enumerate(sorted(resolutions)):
          division[codec][resolution].append(
              # Append the ith resolution if found, else, append the max
              # available resolution.  This would be a valid choice of
              # grouping every time because all the inputs will have a common
              # lowest resolution, that's because they share the same pipeline
              # configuration.  At some point, some input(s) won't be able to
              # scale up in terms of the resolution as the rest, that's when we
              # pick the highest resolution available for it/them.
              codec_division[codec][min(i, len(codec_division[codec]) - 1)])
    
    concat_vid_playlists: List[MediaPlaylist] = []
    for codec, resolution_division in division.items():
      for resolution, vid_playlists in resolution_division.items():
        stream_info: Dict[str, str] = MediaPlaylist._similar_stream_info(
        vid_playlists)
        # Get a unique URI.
        stream_info.update(MediaPlaylist._next_unique_name())
        # NOTE: stream variants don't have a NAME attribute.
        stream_info.pop('NAME')
        # Get the peak and average bandwidth for this codec-resolution pair.
        stream_info.update(MediaPlaylist._get_bandwidth(vid_playlists,
                                                        durations))
        # Get all the codecs that will be inside the new variant stream playlist.
        stream_info['CODECS'] = MediaPlaylist._get_hls_codec(vid_playlists)
        concat_vid_playlist = MediaPlaylist(stream_info)
        concat_vid_playlist.target_duration = MediaPlaylist._max_target_dur(
            vid_playlists)
        for vid_playlist in vid_playlists:
          concat_vid_playlist.content += vid_playlist.content
          # Add a discontinuity after each period.
          concat_vid_playlist.content += '#EXT-X-DISCONTINUITY\n\n'
        concat_vid_playlists.append(concat_vid_playlist)
    
    return concat_vid_playlists


class MasterPlaylist:
  """A class representing a master playlist."""
  
  def __init__(self,
               file_name: Optional[str] = None,
               output_dir: Optional[str] = None,
               packager: Optional[PackagerNode] = None):
    """Given the path to the master playlist file, this method will read that file
    and search for MediaPlaylists inside it, and instantiate a MediaPlaylist object
    for each one found.
    
    The (MediaPlaylist)'s __init__ method will parse each media playlist and change the
    path of media segments to be relative to `output_dir`.
    
    if `file_name` is None, an empty MasterPlaylist object is returned.
    """
    
    self.playlists: List[MediaPlaylist] = []
    self.duration = 0.0
    
    if file_name is None:
      # Do not read. The media playlists will be appended manually.
      return

    # If there is a file to read, we MUST have a bound packager node
    # to match its streams with each media playlist.
    assert packager is not None
    
    # We will fill this map, file names to (OutputStream)s and will give it out
    # to each media playlist we initialize.  Each media playlist then will pick
    # the right output stream for itself.
    streams_map: Dict[str, OutputStream] = {}
    
    # Add a mapping between single segment file names and their output stream.
    streams_map.update({
        outstream.get_single_seg_file().write_end(): outstream
        for outstream in packager.output_streams
      })
    # Add another mapping between the first segment in multi-segment file names
    # and their corresponding output streams.
    streams_map.update({
        outstream.get_media_seg_file()
        .write_end()
        .replace('$Number$', '1'): outstream
        for outstream in packager.output_streams
      })
    
    dir_name = os.path.dirname(file_name)
    
    with open(file_name, 'r') as master_playlist:
      line = master_playlist.readline()
      while line:
        if line.startswith('#EXT-X-MEDIA'):
          stream_info = _extract_attributes(line)
          self.playlists.append(MediaPlaylist(stream_info, dir_name,
                                              output_dir,
                                              streams_map))
        elif line.startswith('#EXT-X-STREAM-INF'):
          stream_info = _extract_attributes(line)
          # Quote the URI to keep consistent,
          # as the URIs in EXT-X-MEDIA are quoted too.
          stream_info['URI'] = _quote(master_playlist.readline().strip())
          self.playlists.append(MediaPlaylist(stream_info, dir_name,
                                              output_dir,
                                              streams_map))
        line = master_playlist.readline()
      # Get the master playlist duration from an arbitrary stream.
      self.duration = self.playlists[-1].duration
  
  def write(self, file: str,
            master_playlist_header: str,
            media_playlist_header: str,
            comment: str) -> None:
    """Writes the master playlist and the nested media playlists in
    the file system.
    """
    
    dir_name = os.path.dirname(file)
    with open(file, 'w') as master_playlist:
      content = master_playlist_header
      content += comment
      # Write #EXT-X-MEDIA media playlists first.
      for media_playlist in self.playlists:
        if media_playlist.stream_info.get('TYPE'):
          media_playlist.write(dir_name, media_playlist_header)
          content += '#EXT-X-MEDIA:' + ','.join(sorted(
              [key + '=' + value for
               key, value in media_playlist.stream_info.items()])) + '\n'
      content += '\n'
      # Then write #EXT-X-STREAM-INF media playlists.
      for media_playlist in self.playlists:
        if media_playlist.stream_info.get('TYPE') is None:
          media_playlist.write(dir_name, media_playlist_header)
          # We don't write the URI in the attributes of a stream
          # variant playlist.  Pop out the URI.
          uri = _unquote(media_playlist.stream_info.pop('URI'))
          content += '#EXT-X-STREAM-INF:' + ','.join(sorted(
              [key + '=' + value for
               key, value in media_playlist.stream_info.items()])) + '\n'
          content += uri + '\n'
      master_playlist.write(content)
  
  @staticmethod
  def extract_headers(file_path: str) -> Tuple[str, str]:
    """Returns the headers for the master and media playlists using the
    master playlist in `file_path` and any media playlist we find inside
    this master playlist.
    """
    
    header = ''
    with open(file_path, 'r') as master_playlist_file:
      line = master_playlist_file.readline()
      # Store each line in header until one of these tags is encountered.
      while line and not line.startswith(('#EXT-X-MEDIA', '#EXT-X-STREAM-INF')):
        # lstrip() will convert empty lines -> '' but will keep non-empty lines unchanged.
        header += line.lstrip()
        line = master_playlist_file.readline()
      else:
        # Use this media playlist to also extract the MediaPlaylist header.
        if line.startswith('#EXT-X-MEDIA'):
          uri = _unquote(_extract_attributes(line)['URI'])
        elif line.startswith('#EXT-X-STREAM-INF'):
          uri = master_playlist_file.readline().strip()
        else:
          raise RuntimeError('No media playlist found in this master playlist')
        master_playlist_dirname = os.path.dirname(file_path)
        media_playlist_path = os.path.join(master_playlist_dirname, uri)
        return header, MediaPlaylist.extract_header(media_playlist_path)
  
  @staticmethod
  def concat_master_playlists(
    master_playlists: List['MasterPlaylist']) -> 'MasterPlaylist':
    
    def var_is_vid(var_playlists: List['MediaPlaylist']) -> bool:
      """This is useful to detect whether the stream variant playlists in some
      master playlist are video playlists or not.  They could be audio playlists
      as #EXT-X-STREAM-INF because there were no video streams associated
      with this input.
      """
      
      # NOTE: Ideally, this check should be performed on all the STREAM-INF playlists
      # for this input, but Shaka-Packager's output guarantees for us that all the
      # STREAM-INFs will be videos or all will be audios in a master playlist.
      assert len(var_playlists), ('There MUST be at least one stream variant'
                                  ' in a master playlist')
      # NOTE: Since we intentionally indexed the list, it must be a non-empty list,
      # i.e. there must be at least one stream variant in each master playlist.
      return isinstance(var_playlists[0].codec, VideoCodec)
    
    # SEMANTICS:
      # xxx_playlist == one MediaPlaylist object.
      # xxx_playlists == multiple MediaPlaylist objects in the same master playlist.
      # all_xxx_playlists == a list of (xxx_playlists).
    all_txt_playlists: List[List['MediaPlaylist']] = []
    all_aud_playlists: List[List['MediaPlaylist']] = []
    all_var_playlists: List[List['MediaPlaylist']] = []
    # The durations will be used to insert a gap in subtitles playlists when
    # there is no applicable media to insert. Will also be used to calculate
    # the average bandwidth.
    durations: List[float] = []
    
    for master_playlist in master_playlists:
      
      txt_playlists: List['MediaPlaylist'] = []
      aud_playlists: List['MediaPlaylist'] = []
      var_playlists: List['MediaPlaylist'] = []
      
      for media_playlist in master_playlist.playlists:
        stream_type = media_playlist.stream_info.get('TYPE', 'STREAM-INF')
        if stream_type == 'SUBTITLES':
          txt_playlists.append(media_playlist)
        elif stream_type == 'AUDIO':
          aud_playlists.append(media_playlist)
        elif stream_type == 'STREAM-INF':
          var_playlists.append(media_playlist)
        else:
          # TODO: We need a case for CLOSED-CAPTIONS(CC).
          raise RuntimeError("TYPE={} is not recognized".format(stream_type))  
      
      all_txt_playlists.append(txt_playlists)
      all_aud_playlists.append(aud_playlists)
      all_var_playlists.append(var_playlists)
      durations.append(master_playlist.duration)
    
    master_hls = MasterPlaylist()
    master_hls.playlists.extend(
        MediaPlaylist.concat_sub(all_txt_playlists, durations))
    
    if all(not var_is_vid(var_pl) for var_pl in all_var_playlists):
      # When the playlist is audio only, each audio is referenced two times,
      # once in an #EXT-X-MEDIA tag and another time in an #EXT-X-STREAM-INF tag.
      # If the user has an audio-only content, the concatenation will go a little
      # bit different to produce the desired output.
      master_hls.playlists.extend(
          MediaPlaylist.concat_aud_only(
            all_aud_playlists,
            all_var_playlists,
            durations))
    else:
      master_hls.playlists.extend(
          MediaPlaylist.concat_aud(all_aud_playlists))
      master_hls.playlists.extend(
          MediaPlaylist.concat_vid(all_var_playlists, durations))
    
    return master_hls


class HLSConcater:
  """A class that serves as an API for the m3u8 concatenation methods."""
  
  def __init__(self,
               sample_master_playlist_path: str,
               output_location: str):
    """Calls MasterPlaylist.extract_headers() and MediaPlaylist.extract_header()
    and store these headers in their classes respectively.
    
    We use the `output_location` to re-evaluate a media segment paths, and also
    write the concatenation result to this `output_location`.
    """
    
    # Extract common master playlist header, this will call
    # MediaPlaylist.extract_header() to extract the common
    # media playlist header as well.
    (self._master_playlist_header,
     self._media_playlist_header) = MasterPlaylist.extract_headers(
          sample_master_playlist_path)
    # Will be used when writing the concatenated playlists.
    self._output_location = output_location
    self._all_master_playlists: List[MasterPlaylist] = []
    
  def add(self, master_playlist_path: str, packager_node: PackagerNode):
    """Adds a master playlist to the HLSConcater object, to be concatented
    in order when HLSConcater.concat() is called.
    """
    
    self._all_master_playlists.append(
        MasterPlaylist(master_playlist_path,
                       self._output_location,
                       packager_node))
  
  def concat_and_write(self, master_playlist_file_name: str, comment: str = ''):
    """Starts concatenating the added master playlists producing one
    final master playlist.
    
    Then, it writes the concatenated playlists in `output_location` parameter
    passed to the constructor.
    """
    
    concated_master_playlist = MasterPlaylist.concat_master_playlists(
        self._all_master_playlists)
    
    if comment:
      comment = '## ' + comment + '\n\n'
    master_playlist_file_name = os.path.join(self._output_location,
                                             master_playlist_file_name)
    concated_master_playlist.write(master_playlist_file_name,
                                   self._master_playlist_header,
                                   self._media_playlist_header,
                                   comment)


def _extract_attributes(line: str) -> Dict[str, str]:
  """Extracts attributes from an m3u8 #EXT-X tag to a python dictionary."""
  
  attributes: Dict[str, str] = {}
  line = line.strip().split(':', 1)[1]
  # For a tighter search, append ',' and search for it in the regex.
  line += ','
  # Search for all KEY=VALUE,
  matches: List[Tuple[str, str]] = re.findall(
      r'([-A-Z]+)=("[^"]*"|[^",]*),', line)
  for key, value in matches:
    attributes[key] = value
  return attributes

def _quote(string: str) -> str:
  """Puts a string in double quotes."""
  
  return '"' + string + '"'

def _unquote(string: str) -> str:
  """Removes the double quotes surrounding a string."""
  
  return string[1:-1]



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
import platform

from . import bitrate_configuration
from . import configuration

from typing import List, Dict, Any, Optional


class InputNotFound(configuration.ConfigError):
  """An error raised when an input stream is not found."""

  def __init__(self, input):
    super().__init__(input.__class__, 'track_num',
                     getattr(input.__class__, 'track_num'))
    self.input = input

  def __str__(self):
    return ('In {}, {} track #{} was'
            ' not found in "{}"').format(self.class_name,
                                         self.input.media_type.value,
                                         self.input.track_num,
                                         self.input.name)

class InputType(enum.Enum):
  FILE = 'file'
  """A track from a file.  Usable only with VOD."""

  LOOPED_FILE = 'looped_file'
  """A track from a file, looped forever by FFmpeg.  Usable only with live.

  Does not support media_type of 'text'.
  """

  WEBCAM = 'webcam'
  """A webcam device.  Usable only with live.

  The device path should be given in the name field.  For example, on Linux,
  this might be /dev/video0.

  Only supports media_type of 'video'.
  """

  MICROPHONE = 'microphone'
  """A microphone device.  Usable only with live.

  The device path should given in the name field.  For example, on Linux, this
  might be "default".

  Only supports media_type of 'audio'.
  """

  EXTERNAL_COMMAND = 'external_command'
  """An external command that generates a stream of audio or video.

  The command should be given in the name field, using shell quoting rules.
  The command should send its generated output to the path in the environment
  variable $SHAKA_STREAMER_EXTERNAL_COMMAND_OUTPUT, which Shaka Streamer set to
  the path to the output pipe.

  May require the user of extra_input_args if FFmpeg can't guess the format or
  framerate.

  Does not support media_type of 'text'.
  """

class MediaType(enum.Enum):
  AUDIO = 'audio'
  VIDEO = 'video'
  TEXT = 'text'


class Input(configuration.Base):
  """An object representing a single input stream to Shaka Streamer."""

  input_type = configuration.Field(InputType, default=InputType.FILE).cast()
  """The type of the input."""

  name = configuration.Field(str, required=True).cast()
  """Name of the input.

  With input_type set to 'file', this is a path to a file name.

  With input_type set to 'looped_file', this is a path to a file name to be
  looped indefinitely in FFmpeg.

  With input_type set to 'webcam', this is which webcam.  On Linux, this is a
  path to the device node for the webcam, such as '/dev/video0'.  On macOS, this
  is a device name, such as 'default'.

  With input_type set to 'external_command', this is an external command that
  generates a stream of audio or video.  The command will be parsed using shell
  quoting rules.  The command should send its generated output to the path in
  the environment variable $SHAKA_STREAMER_EXTERNAL_COMMAND_OUTPUT, which Shaka
  Streamer set to the path to the output pipe.
  """

  extra_input_args = configuration.Field(str, default='').cast()
  """Extra input arguments needed by FFmpeg to understand the input.

  This allows you to take inputs that cannot be understand or detected
  automatically by FFmpeg.

  This string will be parsed using shell quoting rules.
  """

  media_type = configuration.Field(MediaType, required=True).cast()
  """The media type of the input stream."""

  frame_rate = configuration.Field(float).cast()
  """The frame rate of the input stream, in frames per second.

  Only valid for media_type of 'video'.

  Can be auto-detected for some input types, but may be required for others.
  For example, required for input_type of 'external_command'.
  """

  resolution = configuration.Field(
      bitrate_configuration.VideoResolutionName).cast()
  """The name of the input resolution (1080p, etc).

  Only valid for media_type of 'video'.

  Can be auto-detected for some input types, but may be required for others.
  For example, required for input_type of 'external_command'.
  """

  channel_layout = configuration.Field(
      bitrate_configuration.AudioChannelLayoutName).cast()
  """The name of the input channel layout (stereo, surround, etc)."""

  track_num = configuration.Field(int, default=0).cast()
  """The track number of the input.

  The track number is specific to the media_type.  For example, if there is one
  video track and two audio tracks, media_type of 'audio' and track_num of '0'
  indicates the first audio track, not the first track overall in that file.

  If unspecified, track_num will default to 0, meaning the first track matching
  the media_type field will be used.
  """

  is_interlaced = configuration.Field(bool).cast()
  """True if the input video is interlaced.

  Only valid for media_type of 'video'.

  If true, the video will be deinterlaced during transcoding.

  Can be auto-detected for some input types, but may be default to False for
  others.  For example, an input_type of 'external_command', it will default to
  False.
  """

  language = configuration.Field(str).cast()
  """The language of an audio or text stream.

  With input_type set to 'file' or 'looped_file', this will be auto-detected.
  Otherwise, it will default to 'und' (undetermined).
  """

  start_time = configuration.Field(str).cast()
  """The start time of the slice of the input to use.

  Only valid for VOD and with input_type set to 'file'.

  Not supported with media_type of 'text'.
  """

  end_time = configuration.Field(str).cast()
  """The end time of the slice of the input to use.

  Only valid for VOD and with input_type set to 'file'.

  Not supported with media_type of 'text'.
  """

  drm_label = configuration.Field(str).cast()
  """Optional value for a custom DRM label, which defines the encryption key
  applied to the stream. If not provided, the DRM label is derived from stream
  type (video, audio), resolutions, etc. Note that it is case sensitive.

  Applies to 'raw' encryption_mode only."""

  skip_encryption = configuration.Field(int, default=0).cast()
  """If set, no encryption of the stream will be made"""

  # TODO: Figure out why mypy 0.720 and Python 3.7.5 don't correctly deduce the
  # type parameter here if we don't specify it explicitly with brackets after
  # "Field".
  filters = configuration.Field[List[str]](List[str], default=[]).cast()
  """A list of FFmpeg filter strings to add to the transcoding of this input.

  Each filter is a single string.  For example, 'pad=1280:720:20:20'.

  Not supported with media_type of 'text'.
  """

  forced_subtitle = configuration.Field(bool).cast()
  """Indicates that the subtitle is forced.

  Only supported with media_type of 'text'.
  """


  def __init__(self, *args) -> None:
    super().__init__(*args)

    # FIXME: A late import to avoid circular dependency issues between these two
    # modules.
    from . import autodetect

    if not autodetect.is_present(self):
      raise InputNotFound(self)

    def require_field(name: str) -> None:
      """Raise MissingRequiredField if the named field is still missing."""
      if getattr(self, name) is None:
        raise configuration.MissingRequiredField(
            self.__class__, name, getattr(self.__class__, name))

    def disallow_field(name: str, reason: str) -> None:
      """Raise MalformedField if the named field is present."""
      if getattr(self, name):
        raise configuration.MalformedField(
            self.__class__, name, getattr(self.__class__, name), reason)

    if self.media_type == MediaType.VIDEO:
      # These fields are required for video inputs.
      # We will attempt to auto-detect them if possible.
      if self.is_interlaced is None:
        self.is_interlaced = autodetect.get_interlaced(self)

      if self.frame_rate is None:
        self.frame_rate = autodetect.get_frame_rate(self)
      require_field('frame_rate')

      if self.resolution is None:
        self.resolution = autodetect.get_resolution(self)
      require_field('resolution')

    if self.media_type == MediaType.AUDIO:
      if self.language is None:
        self.language = autodetect.get_language(self) or 'und'

      if self.channel_layout is None:
        self.channel_layout = autodetect.get_channel_layout(self)
      require_field('channel_layout')

    if self.media_type == MediaType.TEXT:
      if self.language is None:
        self.language = autodetect.get_language(self) or 'und'
      # Text streams are only supported in plain file inputs.
      if self.input_type != InputType.FILE:
        reason = 'text streams are not supported in input_type "{}"'.format(
            self.input_type.value)
        disallow_field('input_type', reason)
      if self.forced_subtitle is None:
        self.forced_subtitle = autodetect.get_forced_subttitle(self)

      # These fields are not supported with text, because we don't process or
      # transcode it.
      reason = 'not supported with media_type "text"'
      disallow_field('start_time', reason)
      disallow_field('end_time', reason)
      disallow_field('filters', reason)

    if self.input_type != InputType.FILE:
      # These fields are only valid for file inputs.
      reason = 'only valid when input_type is "file"'
      disallow_field('start_time', reason)
      disallow_field('end_time', reason)


  def reset_name(self, pipe_path: str) -> None:
    """Set the name to a pipe path into which this input's contents are fed.
    """

    self.name = pipe_path

  def get_stream_specifier(self) -> str:
    """Get an FFmpeg stream specifier for this input.

    For example, the first video track would be "v:0", and the 3rd text track
    would be "s:2".  Note that all track numbers are per media type in this
    format, not overall track numbers from the input file, and that they are
    indexed starting at 0.

    See also http://ffmpeg.org/ffmpeg.html#Stream-specifiers
    """

    if self.media_type == MediaType.VIDEO:
      return 'v:{}'.format(self.track_num)
    elif self.media_type == MediaType.AUDIO:
      return 'a:{}'.format(self.track_num)
    elif self.media_type == MediaType.TEXT:
      return 's:{}'.format(self.track_num)

    assert False, 'Unrecognized media_type!  This should not happen.'

  def get_input_args(self) -> List[str]:
    """Get any required input arguments for this input.

    These are like hard-coded extra_input_args for certain input types.
    This means users don't have to know much about FFmpeg options to handle
    these common cases.

    Note that for types which support autodetect, these arguments must be
    understood by ffprobe as well as ffmpeg.
    """
    args_matrix: Dict[InputType, Dict[str, List[str]]] = {
        InputType.WEBCAM: {
            'Linux': [
                # Treat the input as a video4linux device, which is how
                # webcams show up on Linux.
                '-f', 'video4linux2',
            ],
            'Darwin': [
                # Webcams on macOS use FFmpeg's avfoundation input format.  With
                # this, you also have to specify an input framerate, unfortunately.
                '-f', 'avfoundation',
                '-framerate', '30',
            ],
            'Windows': [
                # Treat the input as a directshow input device.
                '-f', 'dshow',
            ],
        },
        InputType.MICROPHONE: {
            'Linux': [
                # PulseAudio input device.
                '-f', 'pulse',
            ],
            'Darwin': [
                # AVFoundation also works as an audio input device.
                '-f', 'avfoundation',
            ],
            'Windows': [
                # Directshow also works as an audio input device.
                '-f', 'dshow',
            ],
        },
    }

    args_for_input_type = args_matrix.get(self.input_type)
    # If the input's type wasn't of what interests us.
    if not args_for_input_type:
      return []

    args = args_for_input_type.get(platform.system())
    assert args, '{} is not supported on this platform!'.format(self.input_type.value)

    return args

  def get_resolution(self) -> bitrate_configuration.VideoResolution:
    return bitrate_configuration.VideoResolution.get_value(self.resolution)

  def get_channel_layout(self) -> bitrate_configuration.AudioChannelLayout:
    return bitrate_configuration.AudioChannelLayout.get_value(self.channel_layout)

class SinglePeriod(configuration.Base):
  """An object representing a single period in a multiperiod inputs list."""

  inputs = configuration.Field(List[Input], required=True).cast()

class InputConfig(configuration.Base):
  """An object representing the entire input config to Shaka Streamer."""

  multiperiod_inputs_list = configuration.Field(List[SinglePeriod]).cast()
  """A list of SinglePeriod objects"""

  inputs = configuration.Field(List[Input]).cast()
  """A list of Input objects"""

  def __init__(self, dictionary: Dict[str, Any]):
    """A constructor to check that either inputs or mutliperiod_inputs_list is provided,
    and produce a helpful error message in case both or none are provided.

    We need these checks before passing the input dictionary to the configuration.Base constructor,
    because it does not check for this 'exclusive or-ing' relationship between fields.
    """

    assert isinstance(dictionary, dict), """Malformed Input Config File,
    See some examples at https://github.com/shaka-project/shaka-streamer/tree/main/config_files.
    """

    if (dictionary.get('inputs') is not None
        and dictionary.get('multiperiod_inputs_list') is not None):
      raise configuration.ConflictingFields(
        InputConfig, 'inputs', 'multiperiod_inputs_list')

    # Because these fields are not marked as required at the class level
    # , we need to check ourselves that one of them is provided.
    if not dictionary.get('inputs') and not dictionary.get('multiperiod_inputs_list'):
      raise configuration.MissingRequiredExclusiveFields(
        InputConfig, 'inputs', 'multiperiod_inputs_list')

    super().__init__(dictionary)




# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module that feeds information from two named pipes into shaka-packager."""

import os
import subprocess

from . import input_configuration
from . import node_base
from . import pipeline_configuration

from streamer.bitrate_configuration import AudioCodec, VideoCodec
from streamer.input_configuration import MediaType
from streamer.output_stream import OutputStream
from streamer.pipeline_configuration import EncryptionMode, PipelineConfig
from streamer.util import is_url
from typing import List, Optional, Union, cast

# Alias a few classes to avoid repeating namespaces later.
ManifestFormat = pipeline_configuration.ManifestFormat
StreamingMode = pipeline_configuration.StreamingMode

class SegmentError(Exception):
  """Raise when segment is incompatible with format."""
  pass

def build_path(output_location, sub_path):
  """Handle annoying edge cases with paths for cloud upload.
  If a path has two slashes, GCS will create an intermediate directory named "".
  So we have to be careful in how we construct paths to avoid this.
  """
  # ControllerNode should have already stripped trailing slashes from the output
  # location.

  # Sometimes the segment dir is empty.  This handles that special case.
  if not sub_path:
    return output_location

  if is_url(output_location):
    # Don't use os.path.join, since URLs must use forward slashes and Streamer
    # could be used on Windows.
    return output_location + '/' + sub_path

  return os.path.join(output_location, sub_path)


class PackagerNode(node_base.PolitelyWaitOnFinish):

  def __init__(self,
               pipeline_config: PipelineConfig,
               output_location: str,
               output_streams: List[OutputStream],
               index: int,
               hermetic_packager: Optional[str]) -> None:
    super().__init__()
    self._pipeline_config: PipelineConfig = pipeline_config
    self.output_location: str = output_location
    self._segment_dir: str = build_path(
        output_location, pipeline_config.segment_folder)
    self.output_streams: List[OutputStream] = output_streams
    self._index = index
    # If a hermetic packager is passed, use it.
    self._packager = hermetic_packager or 'packager'

  def start(self) -> None:
    args = [
        self._packager,
    ]

    args += [self._setup_stream(stream) for stream in self.output_streams]

    if self._pipeline_config.quiet:
      args += [
          '--quiet',  # Only output error logs
      ]

    args += [
        # Segment duration given in seconds.
        '--segment_duration', str(self._pipeline_config.segment_size),
        # Signal DolbyVision using the modern supplemental codecs approach
        '--use_dovi_supplemental_codecs',
    ]

    if self._pipeline_config.streaming_mode == StreamingMode.LIVE:
      args += [
          # Number of seconds the user can rewind through backwards.
          '--time_shift_buffer_depth',
          str(self._pipeline_config.availability_window),
          # Number of segments preserved outside the current live window.
          # NOTE: This must not be set below 3, or the first segment in an HLS
          # playlist may become unavailable before the playlist is updated.
          '--preserved_segments_outside_live_window', '3',
          # Number of seconds of content encoded/packaged that is ahead of the
          # live edge.
          '--suggested_presentation_delay',
          str(self._pipeline_config.presentation_delay),
          # Number of seconds between manifest updates.
          '--minimum_update_period',
          str(self._pipeline_config.update_period),
          # Ignore HTTP output failures so that the pipeline doesn't stop if we
          # fail to upload one segment.  Only enable this for live streams,
          # since for VOD, we really should signal that failure to the user.
          '--ignore_http_output_failures',
      ]

    args += self._setup_manifest_format()

    if self._pipeline_config.encryption.enable:
      args += self._setup_encryption()

    stdout = None
    if self._pipeline_config.debug_logs:
      # Log by writing all Packager output to a file.  Unlike the logging
      # system in ffmpeg, this will stop any Packager output from getting to
      # the screen.
      packager_log_file = 'PackagerNode-' + str(self._index) + '.log'
      stdout = open(packager_log_file, 'w')

    self._process: subprocess.Popen = self._create_process(
        args,
        stderr=subprocess.STDOUT,
        stdout=stdout)

  def _setup_stream(self, stream: OutputStream) -> str:
    dict = {
        'in': stream.ipc_pipe.read_end(),
        'stream': stream.type.value,
    }

    if stream.input.skip_encryption:
      dict['skip_encryption'] = str(stream.input.skip_encryption)

    if stream.type == MediaType.AUDIO:
      dict['hls_group_id'] = str(cast(AudioCodec, stream.codec).value)

    if stream.type == MediaType.VIDEO and self._pipeline_config.generate_iframe_playlist:
      dict['iframe_playlist_name'] = 'iframe_' + stream.get_identification() + '.m3u8'

    if stream.input.drm_label:
      dict['drm_label'] = stream.input.drm_label

    if stream.input.forced_subtitle:
      dict['forced_subtitle'] = '1'

    # Note: Shaka Packager will not accept 'und' as a language, but Shaka
    # Player will fill that in if the language metadata is missing from the
    # manifest/playlist.
    if stream.input.language and stream.input.language != 'und':
      dict['language'] = stream.input.language

    if self._pipeline_config.segment_per_file:
      dict['init_segment'] = build_path(
        self._segment_dir,
        stream.get_init_seg_file().write_end())
      dict['segment_template'] = build_path(
        self._segment_dir,
        stream.get_media_seg_file().write_end())
    else:
      dict['output'] = build_path(
        self._segment_dir,
        stream.get_single_seg_file().write_end())

    if stream.is_dash_only():
      dict['dash_only'] = '1'

    # The format of this argument to Shaka Packager is a single string of
    # key=value pairs separated by commas.
    return ','.join(key + '=' + value for key, value in dict.items())

  def _setup_manifest_format(self) -> List[str]:
    args: List[str] = []

    if ManifestFormat.DASH in self._pipeline_config.manifest_format:
      if self._pipeline_config.utc_timings:
        args += [
            '--utc_timings',
            ','.join(timing.scheme_id_uri + '=' +
                     timing.value for timing in self._pipeline_config.utc_timings)
        ]

      if self._pipeline_config.low_latency_dash_mode:
        args += [
            '--low_latency_dash_mode=true',
        ]

      if self._pipeline_config.streaming_mode == StreamingMode.VOD:
        args += [
            '--generate_static_live_mpd',
        ]

      args += [
          # Generate DASH manifest file.
          '--mpd_output',
          build_path(self.output_location, self._pipeline_config.dash_output),
      ]

    if ManifestFormat.HLS in self._pipeline_config.manifest_format:
      if self._pipeline_config.streaming_mode == StreamingMode.LIVE:
        args += [
            '--hls_playlist_type', 'LIVE',
        ]
      else:
        args += [
            '--hls_playlist_type', 'VOD',
        ]

      args += [
          # Generate HLS playlist file(s).
          '--hls_master_playlist_output',
          build_path(self.output_location, self._pipeline_config.hls_output),
      ]

    return args

  def _setup_encryption_keys(self) -> List[str]:
    # Sets up encryption keys for raw encryption mode
    keys = []
    for key in self._pipeline_config.encryption.keys:
      key_str = ''
      if key.label:
        key_str = 'label=' + key.label + ':'
      key_str += 'key_id=' + key.key_id + ':key=' + key.key
      keys.append(key_str)
    return keys

  def _setup_encryption(self) -> List[str]:
    # Sets up encryption of content.

    encryption = self._pipeline_config.encryption

    args = []

    if encryption.encryption_mode == EncryptionMode.WIDEVINE:
      args = [
        '--enable_widevine_encryption',
        '--key_server_url', encryption.key_server_url,
        '--content_id', encryption.content_id,
        '--signer', encryption.signer,
        '--aes_signing_key', encryption.signing_key,
        '--aes_signing_iv', encryption.signing_iv,
      ]
    elif encryption.encryption_mode == EncryptionMode.RAW:
      # raw key encryption mode
      args = [
        '--enable_raw_key_encryption',
        '--keys',
        ','.join(self._setup_encryption_keys()),
      ]
      if encryption.iv:
        args.extend(['--iv', encryption.iv])
      if encryption.pssh:
        args.extend(['--pssh', encryption.pssh])

    # Common arguments
    args.extend([
      '--protection_scheme',
      encryption.protection_scheme.value,
      '--clear_lead', str(encryption.clear_lead),
      '--create_session_keys',
    ])

    if encryption.protection_systems:
      args.extend([
        '--protection_systems', ','.join(
          [p.value for p in encryption.protection_systems]
        )
      ])

    return args



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import enum
import os
import platform
import shlex

from . import bitrate_configuration
from . import configuration

from typing import List


# A randomly-chosen content ID in hex.
RANDOM_CONTENT_ID = base64.b16encode(os.urandom(16)).decode('UTF-8')

# The Widevine UAT server URL.
UAT_SERVER = 'https://license.uat.widevine.com/cenc/getcontentkey/widevine_test'

# Credentials for the Widevine test account.
WIDEVINE_TEST_ACCOUNT = 'widevine_test'
WIDEVINE_TEST_SIGNING_KEY = '1ae8ccd0e7985cc0b6203a55855a1034afc252980e970ca90e5202689f947ab9'
WIDEVINE_TEST_SIGNING_IV = 'd58ce954203b7c9a9a9d467f59839249'

# The default hardware acceleration API to use, per platform.
if platform.system() == 'Linux':
  DEFAULT_HWACCEL_API = 'vaapi'
elif platform.system() == 'Darwin':  # AKA macOS
  DEFAULT_HWACCEL_API = 'videotoolbox'
else:
  DEFAULT_HWACCEL_API = ''


class StreamingMode(enum.Enum):
  LIVE = 'live'
  """Indicates a live stream, which has no end."""

  VOD = 'vod'
  """Indicates a video-on-demand (VOD) stream, which is finite."""

class ManifestFormat(enum.Enum):
  DASH = 'dash'
  HLS = 'hls'

class ProtectionScheme(enum.Enum):
  CENC = 'cenc'
  """AES-128-CTR mode."""

  CBCS = 'cbcs'
  """AES-128-CBC mode with pattern encryption."""

class ProtectionSystem(enum.Enum):
  WIDEVINE = 'Widevine'
  FAIRPLAY = 'FairPlay'
  PLAYREADY = 'PlayReady'
  MARLIN = 'Marlin'
  COMMON = 'CommonSystem'

class EncryptionMode(enum.Enum):
  WIDEVINE = 'widevine'
  """Widevine key server mode"""

  RAW = 'raw'
  """Raw key mode"""

class UtcTimingPair(configuration.Base):
  """An object containing the attributes for a DASH MPD UTCTiming 
  element"""

  # TODO: Use an enum for scheme_id_uri to simplify the config input
  scheme_id_uri = configuration.Field(str).cast()
  """SchemeIdUri attribute to be used for the UTCTiming element"""

  value = configuration.Field(str).cast()
  """Value attribute to be used for the UTCTiming element"""

class RawKeyConfig(configuration.Base):
  """An object representing a list of keys for Raw key encryption"""

  label = configuration.Field(str).cast()
  """An arbitary string or a predefined DRM label like AUDIO, SD, HD, etc.
  If not specified, indicates the default key and key_id."""

  key_id = configuration.Field(configuration.HexString, required=True).cast()
  """A key identifier as a 32-digit hex string"""

  key = configuration.Field(configuration.HexString, required=True).cast()
  """The encryption key to use as a 32-digit hex string"""


class EncryptionConfig(configuration.Base):
  """An object representing the encryption config for Shaka Streamer."""

  enable = configuration.Field(bool, default=False).cast()
  """If true, encryption is enabled.

  Otherwise, all other encryption settings are ignored.
  """

  encryption_mode = configuration.Field(
    EncryptionMode, default=EncryptionMode.WIDEVINE).cast()
  """Encryption mode to use. By default it is widevine but can be changed
  to raw."""

  protection_systems = configuration.Field(List[ProtectionSystem]).cast()
  """Protection Systems to be generated. Supported protection systems include
  Widevine, PlayReady, FairPlay, Marin and CommonSystem.
  """

  pssh = configuration.Field(configuration.HexString).cast()
  """One or more concatenated PSSH boxes in hex string format. If this and
  `protection_systems` is not specified, a v1 common PSSH box will be
  generated.
  
  Applies to 'raw' encryption_mode only.
  """

  iv = configuration.Field(configuration.HexString).cast()
  """IV in hex string format. If not specified, a random IV will be
  generated.
  
  Applies to 'raw' encryption_mode only.
  """

  keys = configuration.Field(List[RawKeyConfig]).cast()
  """A list of encryption keys to use.
  
  Applies to 'raw' encryption_mode only."""

  content_id = configuration.Field(
      configuration.HexString, default=RANDOM_CONTENT_ID).cast()
  """The content ID, in hex.

  If omitted, a random content ID will be chosen for you.
  
  Applies to 'widevine' encryption_mode only.
  """

  key_server_url = configuration.Field(str, default=UAT_SERVER).cast()
  """The URL of your key server.

  This is used to generate an encryption key.  By default, it is Widevine's UAT
  server.
  
  Applies to 'widevine' encryption_mode only.
  """

  signer = configuration.Field(str, default=WIDEVINE_TEST_ACCOUNT).cast()
  """The name of the signer when authenticating to the key server.

  Applies to 'widevine' encryption_mode only.

  Defaults to the Widevine test account.
  """

  signing_key = configuration.Field(
      configuration.HexString, default=WIDEVINE_TEST_SIGNING_KEY).cast()
  """The signing key, in hex, when authenticating to the key server.

  Applies to 'widevine' encryption_mode only.

  Defaults to the Widevine test account's key.
  """

  signing_iv = configuration.Field(
      configuration.HexString, default=WIDEVINE_TEST_SIGNING_IV).cast()
  """The signing IV, in hex, when authenticating to the key server.

  Applies to 'widevine' encryption_mode only.

  Defaults to the Widevine test account's IV.
  """

  protection_scheme = configuration.Field(ProtectionScheme,
                                          default=ProtectionScheme.CENC).cast()
  """The protection scheme (cenc or cbcs) to use when encrypting."""

  clear_lead = configuration.Field(int, default=10).cast()
  """The seconds of unencrypted media at the beginning of the stream."""

  def __init__(self, *args) -> None:
    super().__init__(*args)

    # Don't do any further checks if encryption is disabled
    if not self.enable:
      return

    if self.encryption_mode == EncryptionMode.WIDEVINE:
      field_names = ['keys', 'pssh', 'iv']
      for field_name in field_names:
        if getattr(self, field_name):
          field = getattr(self.__class__, field_name)
          reason = 'cannot be set when encryption_mode is "%s"' % \
                   self.encryption_mode
          raise configuration.MalformedField(
            self.__class__, field_name, field, reason)
    elif self.encryption_mode == EncryptionMode.RAW:
      # Check at least one key has been specified
      if not self.keys:
        field = self.__class__.keys
        reason = 'at least one key must be specified'
        raise configuration.MalformedField(
          self.__class__, 'keys', field, reason)

class PipelineConfig(configuration.Base):
  """An object representing the entire pipeline config for Shaka Streamer."""

  streaming_mode = configuration.Field(StreamingMode, required=True).cast()
  """The streaming mode, which can be either 'vod' or 'live'."""

  quiet = configuration.Field(bool, default=False).cast()
  """If true, reduce the level of output.

  Only errors will be shown in quiet mode.
  """

  debug_logs = configuration.Field(bool, default=False).cast()
  """If true, output simple log files from each node.

  No control is given over log filenames.  Logs are written to the current
  working directory.  We do not yet support log rotation.  This is meant only
  for debugging.
  """

  hwaccel_api = configuration.Field(str, default=DEFAULT_HWACCEL_API).cast()
  """The FFmpeg hardware acceleration API to use with hardware codecs.

  A per-platform default will be chosen if this field is omitted.

  See documentation here: https://trac.ffmpeg.org/wiki/HWAccelIntro
  """

  resolutions = configuration.Field(
      List[bitrate_configuration.VideoResolutionName]).cast()
  """A list of resolution names to encode.

  Any resolution greater than the input resolution will be ignored, to avoid
  upscaling the content.  This also allows you to reuse a pipeline config for
  multiple inputs.

  If not set, it will default to a list of all the (VideoResolutionName)s
  defined in the bitrate configuration.
  """

  channel_layouts = configuration.Field(
      List[bitrate_configuration.AudioChannelLayoutName]).cast()
  """A list of channel layouts to encode.

  Any channel count greater than the input channel count will be ignored.

  If not set, it will default to a list of all the (AudioChannelLayoutName)s
  defined in the bitrate configuration.
  """

  audio_codecs = configuration.Field(
      List[bitrate_configuration.AudioCodec],
      default=[bitrate_configuration.AudioCodec.AAC]).cast()
  """The audio codecs to encode with."""

  video_codecs = configuration.Field(
      List[bitrate_configuration.VideoCodec],
      default=[bitrate_configuration.VideoCodec.H264]).cast()
  """The video codecs to encode with.

  Note that the prefix "hw:" indicates that a hardware encoder should be
  used.
  """

  manifest_format = configuration.Field(List[ManifestFormat],
                                        default=[
                                            ManifestFormat.DASH,
                                            ManifestFormat.HLS,
                                        ]).cast()
  """A list of manifest formats (dash or hls) to create.

  By default, this will create both.
  """

  dash_output = configuration.Field(str, default='dash.mpd').cast()
  """Output filename for the DASH manifest, if created."""

  hls_output = configuration.Field(str, default='hls.m3u8').cast()
  """Output filename for the HLS master playlist, if created."""

  segment_folder = configuration.Field(str, default='').cast()
  """Sub-folder for segment output (or blank for none)."""

  segment_size = configuration.Field(float, default=4).cast()
  """The length of each segment in seconds."""

  segment_per_file = configuration.Field(bool, default=True).cast()
  """If true, force each segment to be in a separate file.

  Must be true for live content.
  """

  generate_iframe_playlist = configuration.Field(bool, default=False).cast()
  """If true, the iFrame playlist will be generated."""

  availability_window = configuration.Field(int, default=300).cast()
  """The number of seconds a segment remains available."""

  presentation_delay = configuration.Field(int, default=30).cast()
  """How far back from the live edge the player should be, in seconds."""

  update_period = configuration.Field(int, default=8).cast()
  """How often the player should fetch a new manifest, in seconds."""

  encryption = configuration.Field(EncryptionConfig,
                                   default=EncryptionConfig({})).cast()
  """Encryption settings."""

  # TODO: Generalize this to low_latency_mode once LL-HLS is supported by Packager
  low_latency_dash_mode = configuration.Field(bool, default=False).cast()
  """If true, stream in low latency mode for DASH."""

  utc_timings = configuration.Field(List[UtcTimingPair]).cast()
  """UTCTiming schemeIdUri and value pairs for the DASH MPD.

  If multiple UTCTiming pairs are provided for redundancy,
  list the pairs in the order of preference.

  Must be set for LL-DASH streaming.
  """


  def __init__(self, *args) -> None:

    # Set the default values of the resolutions and channel_layouts
    # to the values we have in the bitrate configuration.
    # We need the 'type: ignore' here because mypy thinks these variables are lists
    # of VideoResolutionName and AudioChannelLayoutName and not Field variables.
    self.__class__.resolutions.default = list(  # type: ignore
      bitrate_configuration.VideoResolution.keys())
    self.__class__.channel_layouts.default = list(  # type: ignore
      bitrate_configuration.AudioChannelLayout.keys())

    super().__init__(*args)

    if self.streaming_mode == StreamingMode.LIVE and not self.segment_per_file:
      field = self.__class__.segment_per_file
      reason = 'must be true when streaming_mode is "live"'
      raise configuration.MalformedField(
          self.__class__, 'segment_per_file', field, reason)

  def get_resolutions(self) -> List[bitrate_configuration.VideoResolution]:
    VideoResolution = bitrate_configuration.VideoResolution  # alias
    return [VideoResolution.get_value(name) for name in self.resolutions]

  def get_channel_layouts(self) -> List[bitrate_configuration.AudioChannelLayout]:
    AudioChannelLayout = bitrate_configuration.AudioChannelLayout # alias
    return [AudioChannelLayout.get_value(name) for name in self.channel_layouts]



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
import math
import re

from . import configuration
from typing import Dict, Tuple


class BitrateString(configuration.ValidatingType, str):
  """A wrapper that can be used in Field() to require a bitrate string."""

  @staticmethod
  def name() -> str:
    return 'bitrate string'

  @staticmethod
  def validate(value: str) -> None:
    if type(value) is not str:
      raise TypeError()
    if not re.match(r'^[\d\.]+(?:[kM])?$', value):
      raise ValueError('not a bitrate string (e.g. 500k or 7.5M)')


class AudioCodec(enum.Enum):
  AAC = 'aac'
  OPUS = 'opus'
  AC3 = 'ac3'
  EAC3 = 'eac3'
  FLAC = 'flac'

  def is_hardware_accelerated(self) -> bool:
    """Returns True if this codec is hardware accelerated."""
    return False

  def get_ffmpeg_codec_string(self, hwaccel_api: str) -> str:
    """Returns a codec string accepted by FFmpeg for this codec."""
    # FFmpeg warns:
    #   The encoder 'opus' is experimental but experimental codecs are not
    #   enabled, add '-strict -2' if you want to use it. Alternatively use the
    #   non experimental encoder 'libopus'.
    if self == AudioCodec.OPUS:
      return 'libopus'

    return self.value

  def get_output_format(self) -> str:
    """Returns an FFmpeg output format suitable for this codec."""
    # TODO(#31): add support for configurable output format per-codec
    if self in {AudioCodec.OPUS, AudioCodec.AAC, AudioCodec.AC3, AudioCodec.EAC3, AudioCodec.FLAC}:
      return 'mp4'
    else:
      assert False, 'No mapping for output format for codec {}'.format(
          self.value)


class VideoCodec(enum.Enum):

  H264 = 'h264'
  """H264, also known as AVC."""

  VP9 = 'vp9'
  """VP9."""

  AV1 = 'av1'
  """AV1."""
  
  HEVC = 'hevc'
  """HEVC, also known as h.265"""

  def __init__(self, value):
    # Set all the codecs not to be hardware accelerated at the begining.
    self._hw_acc = False

  @classmethod
  def _missing_(cls, value: object) -> 'VideoCodec':
    if isinstance(value, str) and value.startswith('hw:'):
      obj = cls(value[3:])
      # Overwrite the _hw_acc variable for this codec.
      obj._hw_acc = True
      return obj
    return super()._missing_(value)

  def is_hardware_accelerated(self) -> bool:
    """Returns True if this codec is hardware accelerated."""
    return self._hw_acc

  def get_ffmpeg_codec_string(self, hwaccel_api: str) -> str:
    """Returns a codec string accepted by FFmpeg for this codec."""
    if self._hw_acc:
      assert hwaccel_api, 'No hardware encoding support on this platform!'
      return self.value + '_' + hwaccel_api

    return self.value

  def get_output_format(self) -> str:
    """Returns an FFmpeg output format suitable for this codec."""
    # TODO(#31): add support for configurable output format per-codec (mp4 or webm)
    if self in {VideoCodec.H264, VideoCodec.VP9, VideoCodec.HEVC, VideoCodec.AV1}:
      return 'mp4'
    else:
      assert False, 'No mapping for output format for codec {}'.format(
          self.value)


class AudioChannelLayout(configuration.RuntimeMap):

  max_channels = configuration.Field(type=int, required=True).cast()
  """The maximum number of channels in this layout.

  For example, the maximum number of channels for stereo is 2.
  """

  bitrates = configuration.Field(
      Dict[AudioCodec, BitrateString], required=True).cast()
  """A map of audio codecs to the target bitrate for this channel layout.

  For example, in stereo, AAC can have a different bitrate from Opus.

  This value is a string in bits per second, with the suffix 'k' or 'M' for
  kilobits per second or megabits per second.

  For example, this could be '500k' or '7.5M'.
  """

  def _sortable_properties(self) -> Tuple[float]:
    """Return a tuple of properties we can sort on."""
    return (self.max_channels,)


DEFAULT_AUDIO_CHANNEL_LAYOUTS = {
  'mono': AudioChannelLayout({
    'max_channels': 1,
    'bitrates': {
      'aac': '64k',
      'opus': '32k',
      'ac3': '96k',
      'eac3': '48k',
      'flac': '705k',
    },
  }),
  'stereo': AudioChannelLayout({
    'max_channels': 2,
    'bitrates': {
      'aac': '128k',
      'opus': '64k',
      'ac3': '192k',
      'eac3': '96k',
      'flac': '1410k',
    },
  }),
  'surround': AudioChannelLayout({
    'max_channels': 6,
    'bitrates': {
      'aac': '256k',
      'opus': '128k',
      'ac3': '384k',
      'eac3': '192k',
      'flac': '4230k',
    },
  }),
}

class AudioChannelLayoutName(configuration.RuntimeMapKeyValidator):
  """A type which will only allow valid AudioChannelLayout names at runtime."""

  map_class = AudioChannelLayout


class VideoResolution(configuration.RuntimeMap):

  max_width = configuration.Field(type=int, required=True).cast()
  """The maximum width in pixels for this named resolution."""

  max_height = configuration.Field(type=int, required=True).cast()
  """The maximum height in pixels for this named resolution."""

  max_frame_rate = configuration.Field(type=float, default=math.inf).cast()
  """The maximum frame rate in frames per second for this named resolution.

  By default, the max frame rate is unlimited.
  """

  bitrates = configuration.Field(
      Dict[VideoCodec, BitrateString], required=True).cast()
  """A map of video codecs to the target bitrate for this resolution.

  For example, in 1080p, H264 can have a different bitrate from VP9.

  This value is a string in bits per second, with the suffix 'k' or 'M' for
  kilobits per second or megabits per second.

  For example, this could be '500k' or '7.5M'.
  """

  def _sortable_properties(self) -> Tuple[int, int, float]:
    """Return a tuple of properties we can sort on."""
    return (self.max_width, self.max_height, self.max_frame_rate)


class VideoResolutionName(configuration.RuntimeMapKeyValidator):
  """A type which will only allow valid VideoResolution names at runtime."""

  map_class = VideoResolution


# Default bitrates and resolutions are tracked internally at
# go/shaka-streamer-bitrates
# These are common resolutions, and the bitrates per codec are derived from
# internal encoding guidelines.
DEFAULT_VIDEO_RESOLUTIONS = {
  '144p': VideoResolution({
    'max_width': 256,
    'max_height': 144,
    'max_frame_rate': 30,
    'bitrates': {
      'h264': '108k',
      'vp9': '96k',
      'hevc': '96k',
      'av1': '72k',
    },
  }),
  '240p': VideoResolution({
    'max_width': 426,
    'max_height': 240,
    'max_frame_rate': 30,
    'bitrates': {
      'h264': '242k',
      'vp9': '151k',
      'hevc': '151k',
      'av1': '114k',
    },
  }),
  '360p': VideoResolution({
    'max_width': 640,
    'max_height': 360,
    'max_frame_rate': 30,
    'bitrates': {
      'h264': '400k',
      'vp9': '277k',
      'hevc': '277k',
      'av1': '210k',
    },
  }),
  '480p': VideoResolution({  # NTSC analog broadcast TV resolution
    'max_width': 854,
    'max_height': 480,
    'max_frame_rate': 30,
    'bitrates': {
      'h264': '1M',
      'vp9': '512k',
      'hevc': '512k',
      'av1': '389k',
    },
  }),
  '576p': VideoResolution({  # PAL analog broadcast TV resolution
    'max_width': 1024,
    'max_height': 576,
    'max_frame_rate': 30,
    'bitrates': {
      'h264': '1.5M',
      'vp9': '768k',
      'hevc': '768k',
      'av1': '450k',
    },
  }),
  '720p': VideoResolution({
    'max_width': 1280,
    'max_height': 720,
    'max_frame_rate': 30,
    'bitrates': {
      'h264': '2M',
      'vp9': '1M',
      'hevc': '1M',
      'av1': '512k',
    },
  }),
  '720p-hfr': VideoResolution({
    'max_width': 1280,
    'max_height': 720,
    'bitrates': {
      'h264': '3M',
      'vp9': '2M',
      'hevc': '2M',
      'av1': '778k',
    },
  }),
  '1080p': VideoResolution({
    'max_width': 1920,
    'max_height': 1080,
    'max_frame_rate': 30,
    'bitrates': {
      'h264': '4M',
      'vp9': '2M',
      'hevc': '2M',
      'av1': '850k',
    },
  }),
  '1080p-hfr': VideoResolution({
    'max_width': 1920,
    'max_height': 1080,
    'bitrates': {
      'h264': '5M',
      'vp9': '3M',
      'hevc': '3M',
      'av1': '1M',
    },
  }),
  '1440p': VideoResolution({
    'max_width': 2560,
    'max_height': 1440,
    'max_frame_rate': 30,
    'bitrates': {
      'h264': '9M',
      'vp9': '6M',
      'hevc': '6M',
      'av1': '3.5M',
    },
  }),
  '1440p-hfr': VideoResolution({
    'max_width': 2560,
    'max_height': 1440,
    'bitrates': {
      'h264': '14M',
      'vp9': '9M',
      'hevc': '9M',
      'av1': '5M',
    },
  }),
  '4k': VideoResolution({
    'max_width': 4096,
    'max_height': 2160,
    'max_frame_rate': 30,
    'bitrates': {
      'h264': '17M',
      'vp9': '12M',
      'hevc': '12M',
      'av1': '6M',
    },
  }),
  '4k-hfr': VideoResolution({
    'max_width': 4096,
    'max_height': 2160,
    'bitrates': {
      'h264': '25M',
      'vp9': '18M',
      'hevc': '18M',
      'av1': '9M',
    },
  }),
  '8k': VideoResolution({
    'max_width': 8192,
    'max_height': 4320,
    'max_frame_rate': 30,
    'bitrates': {
      'h264': '40M',
      'vp9': '24M',
      'hevc': '24M',
      'av1': '12M',
    },
  }),
  '8k-hfr': VideoResolution({
    'max_width': 8192,
    'max_height': 4320,
    'bitrates': {
      'h264': '60M',
      'vp9': '36M',
      'hevc': '36M',
      'av1': '18M',
    },
  }),
}


class BitrateConfig(configuration.Base):

  audio_channel_layouts = configuration.Field(
      Dict[str, AudioChannelLayout],
      default=DEFAULT_AUDIO_CHANNEL_LAYOUTS).cast()
  """A map of named channel layouts.

  For example, the key would be a name like "stereo", and the value would be an
  object with all the parameters of how stereo audio would be encoded (2
  channels max, bitrates, etc.).
  """

  video_resolutions = configuration.Field(
      Dict[str, VideoResolution], default=DEFAULT_VIDEO_RESOLUTIONS).cast()
  """A map of named resolutions.

  For example, the key would be a name like "1080p", and the value would be an
  object with all the parameters of how 1080p video would be encoded (max size,
  bitrates, etc.)
  """



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import enum
import functools
import re

import typing
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
from typing import Generic, TypeVar, cast


class ConfigError(Exception):
  """A base class for config errors.

  Each subclass provides a meaningful, human-readable string representation in
  English.
  """

  def __init__(self, class_ref, field_name, field):
    self.class_ref = class_ref
    """A reference to the config class that the error refers to."""

    self.class_name = class_ref.__name__
    """The name of the config class that the error refers to."""

    self.field_name = field_name
    """The name of the field that the error refers to."""

    self.field = field
    """The Field metadata object that the error refers to."""


class UnrecognizedField(ConfigError):
  """An error raised when an unrecognized field is encountered in the input."""

  def __str__(self):
    return '{} contains unrecognized field: {}'.format(
        self.class_name, self.field_name)

class WrongType(ConfigError):
  """An error raised when a field in the input has the wrong type."""

  def __str__(self):
    return 'In {}, {} field requires a {}'.format(
        self.class_name, self.field_name, self.field.get_type_name())

class MissingRequiredField(ConfigError):
  """An error raised when a required field is missing from the input."""

  def __str__(self):
    return '{} is missing a required field: {}, a {}'.format(
        self.class_name, self.field_name, self.field.get_type_name())

class MalformedField(ConfigError):
  """An error raised when a field is malformed."""

  def __init__(self, class_ref, field_name, field, reason):
    super().__init__(class_ref, field_name, field)
    self.reason = reason

  def __str__(self):
    return 'In {}, {} field is malformed: {}'.format(
        self.class_name, self.field_name, self.reason)

class ConflictingFields(ConfigError):
  """An error raised when multiple fields are given and only one of them is allowed at a time."""
  
  def __init__(self, class_ref, field1_name, field2_name):
    self.field1_name = field1_name
    self.field1_type = class_ref.__dict__[field1_name].get_type_name()
    self.field2_name = field2_name
    self.field2_type = class_ref.__dict__[field2_name].get_type_name()
    super().__init__(class_ref, field1_name, class_ref.__dict__[field1_name])
  
  def __str__(self):
    return """In {}, these fields are conflicting:
    {} a {}
    and
    {} a {}\n  consider using only one of them.""".format(self.class_name,
                      self.field1_name, self.field1_type,
                      self.field2_name, self.field2_type)
    
class MissingRequiredExclusiveFields(ConfigError):
  """An error raised when one of an exclusively required fields is missing."""
  
  def __init__(self, class_ref, field1_name, field2_name):
    self.field1_name = field1_name
    self.field1_type = class_ref.__dict__[field1_name].get_type_name()
    self.field2_name = field2_name
    self.field2_type = class_ref.__dict__[field2_name].get_type_name()
    super().__init__(class_ref, field1_name, class_ref.__dict__[field1_name])
  
  def __str__(self):
    return """{} is missing a required field. Use exactly one of these fields:
    {} a {}
    or
    {} a {}""".format(self.class_name, self.field1_name, self.field1_type,
                      self.field2_name, self.field2_type)

class ValidatingType(metaclass=abc.ABCMeta):
  """A base wrapper type that validates the input against a limited range.

  Subclasses of this are handled specially by the config system, but this is
  not a real class type that can be constructed.

  Subclasses must implement a static validate() method that takes a value and
  raises TypeError if the input type is wrong or ValueError if it fails
  validation.

  Subclasses must define a "name" attribute with a human-readable name for the
  type.
  """

  @staticmethod
  @abc.abstractmethod
  def validate(value: str) -> None:
    pass

  @staticmethod
  @abc.abstractmethod
  def name() -> str:
    pass


class HexString(ValidatingType, str):
  """A wrapper that can be used in Field() to require a hex string."""

  @staticmethod
  def name() -> str:
    return 'hexadecimal string'

  @staticmethod
  def validate(value):
    if type(value) is not str:
      raise TypeError()
    if not re.match(r'^[a-fA-F0-9]+$', value):
      raise ValueError('not a hexadecimal string')


# A type parameter used by the Generic Field below.
# For example, for a Field with type=str, FieldType would be a string type and
# Type[FieldType] would be the function "str".
FieldType = TypeVar('FieldType')

class Field(Generic[FieldType]):
  # TODO: This class is populated with actual configuration
  # info at runtime. The correctness of the type/value pairs
  # is checked by Base._check_and_convert_type().

  """A container for metadata about individual config fields."""

  def __init__(self,
               type: Optional[Type[FieldType]],
               required: bool = False,
               default: Optional[FieldType] = None) -> None:
    """
    Args:
        type (class or typing module hint): The required type for values of this
            field.
        required (bool): True if this field is required on input.
        default: The default value if the field is not specified.
    """
    subtypes = Field.get_subtypes(type)  # keytype, subtype

    self.type: Optional[Type] = Field.get_underlying_type(type)
    self.keytype: Optional[Type] = subtypes[0]
    self.subtype: Optional[Type] = subtypes[1]
    self.required: bool = required
    self.default: Optional[FieldType] = default

  def get_type_name(self) -> str:
    """Get a human-readable string for the name of self.type."""
    return Field.get_type_name_static(self.type, self.keytype, self.subtype)

  def cast(self) -> FieldType:
    """Called on every Field instance where it is assigned to a configuration
    class property.  For example:

    class FooConfig(configuration.Base):
      name = configuration.Field(type=str, default="Susan").cast()

    At the class level, configuration fields are all Field instances.  At the
    instance level, the configuration Base class constructor sets all the fields
    to the appropriate config value, whose type is the type parameter FieldType.
    In our example above, the "name" Field from the class-level is replaced by a
    string at the instance level.

    Without this cast() method, mypy would be confused, because it would not see
    the metaprogramming that goes on the Base class constructor.  It would think
    that all configuration instances have properties of type "Field".

    To help mypy understand the type of the instance properties without changing
    the class-level property, this cast() method returns "self", but with the
    type info changed to FieldType for mypy's sake.

    In the example above, mypy knows that "name" is not just a Field, but a
    Field[str], based on the "type=str" parameter in the Field constructor.  It
    uses the type parameter FieldType ("str" in this example) to see that cast()
    returns the type "str".

    This allows the metaprogramming approach we take to configuration to remain
    compatible with mypy's static analysis."""
    return cast(FieldType, self)

  @staticmethod
  def get_underlying_type(type: Optional[Type]) -> Optional[Type]:
    """Get the underlying type from a typing module type hint."""

    # In Python 3.8+, you can use typing.get_origin.  It returns None if "type"
    # is something like "str" or "int" instead of "typing.List" or
    # "typing.Dict", so fall back to type itself.
    if hasattr(typing, 'get_origin'):
      return typing.get_origin(type) or type  # type: ignore

    # Before Python 3.8, you can use this undocumented attribute to get the
    # original type.  If this doesn't exist, you are probably dealing with a
    # basic type like "str" or "int".
    if hasattr(type, '__origin__'):
      return type.__origin__ or type  # type: ignore

    return type

  @staticmethod
  def get_subtypes(
      type: Optional[Type]) -> Tuple[Optional[Type], Optional[Type]]:
    """For Dict hints, returns (keytype, valuetype).

    For List hints, returns (None, valuetype).

    For everything else, returns (None, None)."""

    # In Python 3.8+, you can use typing.get_args.  It returns () if "type"
    # is something like "str" or "int" instead of "typing.List" or
    # "typing.Dict".
    if hasattr(typing, 'get_args'):
      args = typing.get_args(type)
    elif hasattr(type, '__args__'):
      # Before Python 3.8, you can use this undocumented attribute to get the
      # type parameters.  If this doesn't exist, you are probably dealing with a
      # basic type like "str" or "int".
      args = getattr(type, '__args__')
    else:
      args = ()

    underlying = Field.get_underlying_type(type)
    if underlying is dict:
      return cast(Tuple[Optional[Type], Optional[Type]], args)
    if underlying is list:
      return (None, args[0])
    return (None, None)

  @staticmethod
  def get_type_name_static(type: Optional[Type],
                           keytype: Optional[Type],
                           subtype: Optional[Type]) -> str:
    """Get a human-readable string for the name of type."""

    # Make these special cases a little more readable.
    if type is str:
      # Call it a string, not a "str".
      return 'string'
    elif type is list:
      # Mention the subtype.
      return 'list of {}'.format(
          Field.get_type_name_static(subtype, None, None))
    elif type is dict:
      # Mention the subtype.
      return 'dictionary of {} to {}'.format(
          Field.get_type_name_static(keytype, None, None),
          Field.get_type_name_static(subtype, None, None))
    elif type is None:
      # This is only here to allow generic handling of UnrecognizedField errors.
      return 'None'
    elif issubclass(type, enum.Enum):
      # Get the list of valid options as quoted strings.
      options = [repr(str(member.value)) for member in type]
      return '{} (one of {})'.format(type.__name__, ', '.join(options))
    elif issubclass(type, ValidatingType):
      return type.name()

    # Otherwise, return the name of the type.
    return type.__name__


class Base(object):
  """A base class for config objects.

  This will handle all validation, type-checking, defaults, and extraction of
  values from an input dictionary.

  Subclasses must define class-level Field objects defining their fields.
  The base class does the rest.
  """

  def __init__(self, dictionary: Dict[str, Any]) -> None:
    """Ingests, type-checks, and validates the input dictionary."""

    # Collect all the config fields for this type.
    config_fields = {}
    for key, field in self.__class__.__dict__.items():
      if isinstance(field, Field):
        config_fields[key] = field

    for key, value in dictionary.items():
      field = config_fields.get(key)

      # Look for unrecognized fields in the input.
      if not field:
        raise UnrecognizedField(self.__class__, key, Field(None))

      # Check types on recognized fields.
      value = self._check_and_convert_type(field, key, value)

      # Assign the value to self.
      setattr(self, key, value)

    # Look for missing fields.
    for key, field in config_fields.items():
      if not key in dictionary:
        # If it's required, raise an error.
        if field.required:
          raise MissingRequiredField(self.__class__, key, field)

        # Otherwise, assign a default.
        setattr(self, key, field.default)

  def _check_and_convert_type(self,
                              field: Field,
                              key: str,
                              value: Any) -> Any:
    """Check the type of |value| and convert it as necessary.

    Args:
        field (Field): The field definition.
        key (str): The name of the field.
        value: The value to be checked and converted.

    This checks the type of |value| according to |field| and performs any
    necessary conversions to the target type.  This is where any type-specific
    logic or special cases are handled.

    Note that automatic type coercion is avoided.  We wouldn't want a string
    containing the word "False" coerced to boolean True.
    """

    # For fields containing other config objects, specially check and convert
    # them.
    assert field.type is not None, 'No type info for Field {}'.format(key)
    if issubclass(field.type, Base):
      # A config object at this stage should be a dictionary.
      if not isinstance(value, dict):
        raise WrongType(self.__class__, key, field)

      # Let the type of the sub-object validate its contents.
      sub_object = field.type(value)

      # Return this typed object, which will be assigned to self.
      return sub_object

    # For lists, check the type of the value itself, then check the subtype of
    # the list items.
    if field.type is list:
      if not isinstance(value, list):
        raise WrongType(self.__class__, key, field)

      subfield = Field(field.subtype)
      try:
        return [self._check_and_convert_type(subfield, key, v) for v in value]
      except WrongType as e:
        # If conversion raises WrongType on the subfield, raise a WrongType
        # error on this field instead, so that the printed error makes sense.
        if e.field == subfield:
          raise WrongType(self.__class__, key, field) from None
        else:
          raise

    # For dictionaries, check the type of the value itself, then check the
    # type of the dictionary keys, then the type of the dictionary values.
    if field.type is dict:
      if not isinstance(value, dict):
        raise WrongType(self.__class__, key, field)

      keyfield = Field(field.keytype)
      subfield = Field(field.subtype)
      try:
        converted_dict = {}
        for subkey, subvalue in value.items():
          subkey = self._check_and_convert_type(keyfield, key, subkey)
          subvalue = self._check_and_convert_type(subfield, key, subvalue)
          converted_dict[subkey] = subvalue
        return converted_dict
      except WrongType as e:
        # If conversion raises WrongType on the subfield or keyfield, raise a
        # WrongType error on this field instead, so that the printed error
        # makes sense.
        if e.field == keyfield or e.field == subfield:
          raise WrongType(self.__class__, key, field) from None
        else:
          raise

    # For enums, try to case the value to the enum type, and raise a WrontType
    # error if this fails.
    if issubclass(field.type, enum.Enum):
      try:
        return field.type(value)
      except ValueError:
        raise WrongType(self.__class__, key, field) from None

    # A float should be permissive and accept an int, as well.
    if field.type is float:
      if not isinstance(value, (float, int)):
        raise WrongType(self.__class__, key, field)
      return value

    # For strings, accept bools, ints, and floats, too.  For example, "true",
    # "7", and "10.2" are all valid strings, but will come out of the YAML
    # parser as a bool, and int, and a float, respectively.  Convert these to
    # string.
    if field.type is str:
      if isinstance(value, (bool, float, int, str)):
        return str(value)
      raise WrongType(self.__class__, key, field)

    # These aren't true types, but validation classes for specific types of
    # limited input.  Run the validate() method to check the input.
    if issubclass(field.type, ValidatingType):
      try:
        field.type.validate(value)
      except TypeError:
        raise WrongType(self.__class__, key, field) from None
      except ValueError as e:
        raise MalformedField(self.__class__, key, field, str(e)) from None
      return value

    # For all other types, just do a basic type check and return the original
    # value.
    if not isinstance(value, field.type):
      raise WrongType(self.__class__, key, field)
    return value


# A type parameter used by the Generic RuntimeMap below.  Concrete types used
# here must inherit from the RuntimeMap base class.
RuntimeMapSubclass = TypeVar('RuntimeMapSubclass', bound='RuntimeMap')

# This decorator makes it so that we only have to implement __eq__ and __lt__
# to make the instances sortable.  These magic methods in turn depend on
# _sortable_properties, which subclasses must implement.
@functools.total_ordering
class RuntimeMap(Generic[RuntimeMapSubclass], Base):
  """Maintains a map of keys to specific instances from the config file.

  This means a Field can have its type defined before the valid keys/values of
  that type are known.  For example, this is used for resolutions, which are
  defined by a config file.

  After calling set_map on the subclass, the get_value method can be used to
  look up a value from its key.  For example, after setting the map to
  {'foo': 'bar'}, 'foo' becomes the only valid key.  Passing the key 'foo' to
  the get_value then results in the value 'bar' being returned.
  """

  _map: Dict[str, RuntimeMapSubclass] = {}


  def get_key(self) -> str:
    """This defines the synthetic 'get_key' property which will be attached to
    instances later inside set_map.  This definition is only for the sake of
    mypy's type analysis, and will never be called."""

    raise RuntimeError('Synthetic get_key is missing on RuntimeMapSubclass!')

  # NOTE: mypy 0.770 won't accept the @abc.abstractmethod decorator here.  The
  # error is "Only concrete class can be given where
  # Type[RuntimeMap[RuntimeMapSubclass]] is expected."  Removing the decorator
  # makes this class technically concrete.  We still have the RuntimeError below
  # if a subclass doesn't implement this required method.
  def _sortable_properties(self) -> Tuple:
    """Return a tuple of sortable properties.  Implemented by subclasses."""
    raise RuntimeError('_sortable_properties missing on RuntimeMapSubclass!')

  def __eq__(self, other: Any) -> bool:
    return self._sortable_properties() == other._sortable_properties()

  def __lt__(self, other: Any) -> bool:
    return self._sortable_properties() < other._sortable_properties()

  def __hash__(self) -> int:
      return super().__hash__()

  @classmethod
  def set_map(cls,
              map: Dict[str, RuntimeMapSubclass]) -> None:
    """Set the map of valid values for this class."""

    assert cls != RuntimeMap, 'Do not use the base class directly!'
    cls._map = map

    # Synthesize a method on each value to allow the key to be recovered.
    # Use a default parameter in the lambda to effectively bind the parameter,
    # as described here: https://stackoverflow.com/a/19837683
    # Not doing this causes the lambda to always return the key from the final
    # iteration of the loop (a problem familiar to many JavaScript developers).
    for key, value in map.items():
      setattr(value, 'get_key', lambda bound_key=key: bound_key)

  @classmethod
  def get_value(cls, key: str) -> RuntimeMapSubclass:
    """Get a valid value by its key."""

    assert cls != RuntimeMap, 'Do not use the base class directly!'
    try:
      return cls._map[key]
    except KeyError:
      raise ValueError(
          '{} is not a valid {}'.format(key, cls.__name__)) from None

  @classmethod
  def keys(cls):
    """This allows the config system to print the list of allowed strings."""
    return cls._map.keys()

  @classmethod
  def sorted_values(cls) -> List[RuntimeMapSubclass]:
    return sorted(cls._map.values())


class RuntimeMapKeyValidator(ValidatingType, str):
  """A validator that only allows the valid keys for a certain RuntimeMap
  subclass.

  A RuntimeMap subclass should be paired with a RuntimeMapKeyValidator subclass.
  The RuntimeMapKeyValidator subclass should have a "map_class" variable which
  points to the RuntimeMap subclass."""

  """Must be provided by subclasses and point to the matching RuntimeMap
  subclass."""
  map_class: Type[RuntimeMap] = None  # type: ignore

  @classmethod
  def name(cls) -> str:
    options = [repr(str(key)) for key in cls.map_class.keys()]
    return '{} name (one of {})'.format(
        cls.map_class.__name__, ', '.join(options))

  @classmethod
  def validate(cls, key):
    if type(key) is not str:
      raise TypeError()

    if key not in cls.map_class.keys():
      raise ValueError(
          '{} is not a valid {}'.format(key, cls.name()))




# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base classes for nodes."""

import abc
import enum
import os
import shlex
import subprocess
import sys
import threading
import time
import traceback

from . import node_base
from typing import Any, Dict, IO, List, Optional, Union

class ProcessStatus(enum.Enum):
  # Use number values so we can sort based on value.

  Finished = 0
  """The node has completed its task and shut down."""

  Running = 1
  """The node is still running."""

  Errored = 2
  """The node has failed."""


class NodeBase(object):
  """A base class for nodes that run a single subprocess."""

  @abc.abstractmethod
  def __init__(self) -> None:
    self._process: Optional[subprocess.Popen] = None

  def __del__(self) -> None:
    # If the process isn't stopped by now, stop it here.  It is preferable to
    # explicitly call stop().
    self.stop(None)

  @abc.abstractmethod
  def start(self):
    """Start the subprocess.

    Should be overridden by the subclass to construct a command line, call
    self._create_process, and assign the result to self._process.
    """
    pass

  def _create_process(self,
                      args: Union[str, List[str]],
                      env: Dict[str, str] = {},
                      merge_env: bool = True,
                      stdout: Union[int, IO[Any], None] = None,
                      stderr: Union[int, IO[Any], None] = None,
                      shell: bool = False, **kwargs) -> subprocess.Popen:
    """A central point to create subprocesses, so that we can debug the
    command-line arguments.

    Args:
      args: An array of strings if shell is False, or a single string is shell
            is True; the command line of the subprocess.
      env: A dictionary of environment variables to pass to the subprocess.
      merge_env: If true, merge env with the parent process environment.
      shell: If true, args must be a single string, which will be executed as a
             shell command.
    Returns:
      The Popen object of the subprocess.
    """
    if merge_env:
      child_env = os.environ.copy()
      child_env.update(env)
    else:
      child_env = env

    # Print arguments formatted as output from bash -x would be.
    # This makes it easy to see the arguments and easy to copy/paste them for
    # debugging in a shell.
    if shell:
      assert isinstance(args, str)
      print('+ ' + args)
    else:
      assert type(args) is list
      print('+ ' + ' '.join([shlex.quote(arg) for arg in args]))


    return subprocess.Popen(args,
                            env=child_env,
                            stdin=subprocess.DEVNULL,
                            stdout=stdout, stderr=stderr,
                            shell=shell, **kwargs)

  def check_status(self) -> ProcessStatus:
    """Returns the current ProcessStatus of the node."""
    if not self._process:
      raise ValueError('Must have a process to check')

    self._process.poll()
    if self._process.returncode is None:
      return ProcessStatus.Running

    if self._process.returncode == 0:
      return ProcessStatus.Finished
    else:
      return ProcessStatus.Errored

  def stop(self, status: Optional[ProcessStatus]) -> None:
    """Stop the subprocess if it's still running."""
    if self._process:
      # Slightly more polite than kill.  Try this first.
      self._process.terminate()

      if self.check_status() == ProcessStatus.Running:
        # If it's not dead yet, wait 1 second.
        time.sleep(1)

      if self.check_status() == ProcessStatus.Running:
        # If it's still not dead, use kill.
        self._process.kill()
        # Wait for the process to die and read its exit code.  There is no way
        # to ignore a kill signal, so this will happen quickly.  If we don't do
        # this, it can create a zombie process.
        self._process.wait()

class PolitelyWaitOnFinish(node_base.NodeBase):
  """A mixin that makes stop() wait for the subprocess if status is Finished.

  This is as opposed to the base class behavior, in which stop() forces
  the subprocesses of a node to terminate.
  """

  def stop(self, status: Optional[ProcessStatus]) -> None:
    if self._process and status == ProcessStatus.Finished:
      try:
        print('Waiting for', self.__class__.__name__)
        self._process.wait(timeout=300)  # 5m timeout
      except subprocess.TimeoutExpired:
        traceback.print_exc()  # print the exception
        # Fall through.

    super().stop(status)

class ThreadedNodeBase(NodeBase):
  """A base class for nodes that run a thread.

  The thread repeats some callback in a background thread.
  """

  _thread: Optional[threading.Thread]

  def __init__(self, thread_name: str, continue_on_exception: bool, sleep_time: float):
    super().__init__()
    self._status = ProcessStatus.Finished
    self._thread_name = thread_name
    self._thread = None
    self._continue_on_exception = continue_on_exception
    self._sleep_time = sleep_time
    self._sleep_waker_event = threading.Event()

  def _thread_main(self) -> None:
    while self._status == ProcessStatus.Running:
      try:
        self._thread_single_pass()
      except:
        print('Exception in', self._thread_name, '-', sys.exc_info())

        if self._continue_on_exception:
          print(self.__class__.__name__+": 'Continuing.'")
        else:
          print(self.__class__.__name__+": 'Quitting.'")
          self._status = ProcessStatus.Errored
          return

      # Wait a little bit before performing the next pass.
      self._sleep_waker_event.wait(self._sleep_time)

  @abc.abstractmethod
  def _thread_single_pass(self) -> None:
    """Runs a single step of the thread loop.

    This is implemented by subclasses to do whatever it is they do.  It will be
    called repeatedly by the base class from the node's background thread.  If
    this method raises an exception, the behavior depends on the
    continue_on_exception argument in the constructor.  If
    continue_on_exception is true, the the thread will continue.  Otherwise, an
    exception will stop the thread and therefore the node.
    """
    pass

  def start(self) -> None:
    self._status = ProcessStatus.Running
    self._thread = threading.Thread(target=self._thread_main, name=self._thread_name)
    self._thread.start()

  def stop(self, status: Optional[ProcessStatus]) -> None:
    self._status = ProcessStatus.Finished
    # If the thread was sleeping, wake it up.
    self._sleep_waker_event.set()
    if self._thread:
      self._thread.join()

  def check_status(self) -> ProcessStatus:
    return self._status



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module to contain auto-detection logic; based on ffprobe."""

import shlex
import subprocess
import time

from streamer.bitrate_configuration import (AudioChannelLayout, AudioChannelLayoutName,
                                            VideoResolution, VideoResolutionName)
from streamer.input_configuration import Input, InputType
from typing import Optional, List

# These cannot be probed by ffprobe.
TYPES_WE_CANT_PROBE = [
  InputType.EXTERNAL_COMMAND,
]

# This module level variable might be set by the controller node
# if the user chooses to use the shaka streamer bundled binaries.
hermetic_ffprobe: Optional[str] = None

def _probe(input: Input, field: str) -> Optional[str]:
  """Autodetect some feature of the input, if possible, using ffprobe.

  Args:
    input (Input): An input object from input_configuration.
    field (str): A field to pass to ffprobe's -show_entries option.

  Returns:
    The requested field from ffprobe as a string, or None if this fails.
  """

  if input.input_type in TYPES_WE_CANT_PROBE:
    # Not supported for this type.
    return None

  args: List[str] = [
      # Probe this input file
      hermetic_ffprobe or 'ffprobe',
      input.name,
  ]

  # Add any required input arguments for this input type
  args += input.get_input_args()

  args += [
      # Specifically, this stream
      '-select_streams', input.get_stream_specifier(),
      # Show the needed metadata only
      '-show_entries', field,
      # Print the metadata in a compact form, which is easier to parse
      '-of', 'compact=p=0:nk=1'
  ]

  print('+ ' + ' '.join([shlex.quote(arg) for arg in args]))

  output_bytes: bytes = subprocess.check_output(args, stderr=subprocess.DEVNULL)
  # The output is either some probe information or just a blank line.
  output_string: str = output_bytes.decode('utf-8').strip()
  # With certain container formats, ffprobe returns a duplicate
  # output and some empty lines in between. Issue #119
  output_string = output_string.split('\n')[0]
  # After stripping the newline, we can fall back to None if it's empty.
  probe_output: Optional[str] = output_string or None

  # Webcams on Linux seem to behave badly if the device is rapidly opened and
  # closed.  Therefore, sleep for 1 second after a webcam probe.
  if input.input_type == InputType.WEBCAM:
    time.sleep(1)

  return probe_output

def is_present(input: Input) -> bool:
  """Returns true if the stream for this input is indeed found.

  If we can't probe this input type, assume it is present."""

  return bool(_probe(input, 'stream=index') or
              input.input_type in TYPES_WE_CANT_PROBE)

def get_language(input: Input) -> Optional[str]:
  """Returns the autodetected the language of the input."""
  return _probe(input, 'stream_tags=language')

def get_interlaced(input: Input) -> bool:
  """Returns True if we detect that the input is interlaced."""
  interlaced_string = _probe(input, 'stream=field_order')

  # These constants represent the order of the fields (2 fields per frame) of
  # different types of interlaced video.  They can be found in
  # https://www.ffmpeg.org/ffmpeg-codecs.html under the description of the
  # field_order option.  Anything else (including None) should be considered
  # progressive (non-interlaced) video.
  return interlaced_string in [
    'tt',
    'bb',
    'tb',
    'bt',
  ]

def get_frame_rate(input: Input) -> Optional[float]:
  """Returns the autodetected frame rate of the input."""

  frame_rate_string = _probe(input, 'stream=avg_frame_rate')
  if frame_rate_string is None:
    return None

  # This string is the framerate in the form of a fraction, such as '24/1' or
  # '30000/1001'.  Occasionally, there is a pipe after the framerate, such as
  # '32700/1091|'.  We must split it into pieces and do the division to get a
  # float.
  fraction = frame_rate_string.rstrip('|').split('/')
  if len(fraction) == 1:
    frame_rate = float(fraction[0])
  else:
    frame_rate = float(fraction[0]) / float(fraction[1])

  # The detected frame rate for interlaced content is twice what it should be.
  # It's actually the field rate, where it takes two interlaced fields to make
  # a frame.  Because we have to know if it's interlaced already, we must
  # assert that is_interlaced has been set before now.
  assert input.is_interlaced is not None
  if input.is_interlaced:
    frame_rate /= 2.0

  return frame_rate

def get_resolution(input: Input) -> Optional[VideoResolutionName]:
  """Returns the autodetected resolution of the input."""

  resolution_string = _probe(input, 'stream=width,height')
  if resolution_string is None:
    return None

  # This is the resolution of the video in the form of 'WIDTH|HEIGHT'.  For
  # example, '1920|1080'.  Occasionally, there is a pipe after the resolution, 
  # such as '1920|1080|'.  We have to split up width and height and match that
  # to a named resolution.
  width_string, height_string = resolution_string.rstrip('|').split('|')
  width, height = int(width_string), int(height_string)

  for bucket in VideoResolution.sorted_values():
    # The first bucket this fits into is the one.
    if (width <= bucket.max_width and height <= bucket.max_height and
        input.frame_rate <= bucket.max_frame_rate):
      return bucket.get_key()

  return None

def get_channel_layout(input: Input) -> Optional[AudioChannelLayoutName]:
  """Returns the autodetected channel count of the input."""

  channel_count_string = _probe(input, 'stream=channels')
  if channel_count_string is None:
    return None

  channel_count = int(channel_count_string)
  for bucket in AudioChannelLayout.sorted_values():
    if channel_count <= bucket.max_channels:
      return bucket.get_key()

  return None

def get_forced_subttitle(input: Input) -> bool:
  """Returns the forced subtitle value of the input."""

  forced_subttitle_string = _probe(input, 'disposition=forced')

  if forced_subttitle_string is None:
    return False

  return bool(forced_subttitle_string)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Concatenates inputs into periods by creating a master DASH/HLS file."""

import os
import re
import time
from typing import List
from xml.etree import ElementTree
from streamer import __version__
from streamer.node_base import ProcessStatus, ThreadedNodeBase
from streamer.packager_node import PackagerNode
from streamer.pipeline_configuration import PipelineConfig, ManifestFormat
from streamer.output_stream import AudioOutputStream, VideoOutputStream
from streamer.m3u8_concater import HLSConcater


class PeriodConcatNode(ThreadedNodeBase):
  """A node that concatenates multiple DASH manifests and/or HLS playlists
  when the input is a multiperiod_inputs_list and the output is to the local the system.
  """

  def __init__(self,
               pipeline_config: PipelineConfig,
               packager_nodes: List[PackagerNode],
               output_location: str) -> None:
    """Stores all relevant information needed for the period concatenation."""
    super().__init__(thread_name='periodconcat', continue_on_exception=False, sleep_time=3)
    self._pipeline_config = pipeline_config
    self._output_location = output_location
    self._packager_nodes: List[PackagerNode] = packager_nodes
    self._concat_will_fail = False
    
    # know whether the first period has video and audio or not.
    fp_has_vid, fp_has_aud = False, False
    for output_stream in packager_nodes[0].output_streams:
      if isinstance(output_stream, VideoOutputStream):
        fp_has_vid = True
      elif isinstance(output_stream, AudioOutputStream):
        fp_has_aud = True
    
    for i, packager_node in enumerate(self._packager_nodes):
      has_vid, has_aud = False, False
      for output_stream in packager_node.output_streams:
        if isinstance(output_stream, VideoOutputStream):
          has_vid = True
        elif isinstance(output_stream, AudioOutputStream):
          has_aud = True
      if has_vid != fp_has_vid or has_aud != fp_has_aud:
        self._concat_will_fail = True
        print("\nWARNING: Stopping period concatenation.")
        print("Period#{} has {}video and has {}audio while Period#1 "
              "has {}video and has {}audio.".format(i + 1, 
                                                    "" if has_vid else "no ",
                                                    "" if has_aud else "no ",
                                                    "" if fp_has_vid else "no ",
                                                    "" if fp_has_aud else "no "))
        print("\nHINT:\n\tBe sure that either all the periods have video or all do not,\n"
              "\tand all the periods have audio or all do not, i.e. don't mix videoless\n"
              "\tperiods with other periods that have video.\n"
              "\tThis is necessary for the concatenation to be performed successfully.\n")
        time.sleep(5)
        break
  
  def _thread_single_pass(self) -> None:
    """Watches all the PackagerNode(s), if at least one of them is running it skips this
    _thread_single_pass, if all of them are finished, it starts period concatenation, if one of
    them is errored, it raises a RuntimeError.
    """

    for i, packager_node in enumerate(self._packager_nodes):
      status = packager_node.check_status()
      if status == ProcessStatus.Running:
        return
      elif status == ProcessStatus.Errored:
        raise RuntimeError(
          'Concatenation is stopped due '
          'to an error in PackagerNode#{}.'.format(i + 1))
    
    if self._concat_will_fail:
      raise RuntimeError('Unable to concatenate the inputs.')
    
    if ManifestFormat.DASH in self._pipeline_config.manifest_format:
      self._dash_concat()
    
    if ManifestFormat.HLS in self._pipeline_config.manifest_format:
      self._hls_concat()

    self._status = ProcessStatus.Finished

  def _dash_concat(self) -> None:
    """Concatenates multiple single-period DASH manifests into one multi-period DASH manifest."""

    def find(elem: ElementTree.Element, *args: str) -> ElementTree.Element:
      """A better interface for the Element.find() method.
      Use it only if it is guaranteed that the element we are searching for is inside,
      Otherwise it will raise an AssertionError."""

      full_path = '/'.join(['shaka-live:' + tag for tag in args])
      child_elem =  elem.find(full_path, {'shaka-live': default_dash_namespace})

      # elem.find() returns either an ElementTree.Element or None.
      assert child_elem is not None, 'Unable to find: {} using the namespace: {}'.format(
        full_path, default_dash_namespace)
      return child_elem

    # Periods that are going to be collected from different MPD files.
    periods: List[ElementTree.Element] = []

    # Get the root of an MPD file that we will concatenate periods into.
    concat_mpd = ElementTree.ElementTree(file=os.path.join(
      self._packager_nodes[0].output_location,
      self._pipeline_config.dash_output)).getroot()

    # Get the default namespace.
    namespace_matches = re.search(r'\{([^}]*)\}', concat_mpd.tag)
    assert namespace_matches is not None, 'Unable to find the default namespace.'
    default_dash_namespace = namespace_matches.group(1)

    # Remove the 'mediaPresentationDuration' attribute.
    concat_mpd.attrib.pop('mediaPresentationDuration')
    # Remove the Period element in that MPD element.
    concat_mpd.remove(find(concat_mpd, 'Period'))

    for packager_node in self._packager_nodes:
      mpd = ElementTree.ElementTree(file=os.path.join(
        packager_node.output_location,
        self._pipeline_config.dash_output)).getroot()
      period = find(mpd, 'Period')
      period.attrib['duration'] = mpd.attrib['mediaPresentationDuration']

      # A BaseURL that will have the relative path to media file.
      base_url = ElementTree.Element('{{{}}}BaseURL'.format(default_dash_namespace))
      base_url.text = os.path.relpath(packager_node.output_location, self._output_location) + '/'
      period.insert(0, base_url)

      periods.append(period)

    # Add the periods collected from all the files.
    concat_mpd.extend(periods)

    # Write the period concat to the output_location.
    with open(os.path.join(
        self._output_location,
        self._pipeline_config.dash_output), 'w') as master_dash:

      contents = "<?xml version='1.0' encoding='UTF-8'?>\n"
      # TODO: Add Shaka-Packager version to this xml comment.
      contents += "<!--Generated with https://github.com/shaka-project/shaka-packager -->\n"
      contents += "<!--Made Multi-Period with https://github.com/shaka-project/shaka-streamer version {} -->\n".format(__version__)

      # xml.ElementTree replaces the default namespace with 'ns0'.
      # Register the DASH namespace back as the default namespace before converting to string.
      ElementTree.register_namespace('', default_dash_namespace)
      
      # xml.etree.ElementTree already has an ElementTree().write() method,
      # but it won't allow putting comments at the begining of the file.
      contents += ElementTree.tostring(element=concat_mpd, encoding='unicode')
      master_dash.write(contents)
  
  def _hls_concat(self) -> None:
    """Concatenates multiple HLS playlists using #EXT-X-DISCONTINUITY."""
    
    # Initialize the HLS concater with a sample Master HLS playlist and
    # the output location of the concatenated playlists.
    first_hls_playlist = os.path.join(self._packager_nodes[0].output_location,
                                      self._pipeline_config.hls_output)
    # NOTE: Media files' segments location will be relative to this
    # self._output_location we pass to the constructor.
    hls_concater = HLSConcater(first_hls_playlist, self._output_location)
    
    for packager_node in self._packager_nodes:
      hls_playlist = os.path.join(packager_node.output_location,
                                  self._pipeline_config.hls_output)
      hls_concater.add(hls_playlist, packager_node)
    
    # Start the period concatenation and write the output in the output location
    # passed to the HLSConcater at the construction time.
    hls_concater.concat_and_write(
        self._pipeline_config.hls_output,
        'Concatenated with https://github.com/shaka-project/shaka-streamer'
        ' version {}'.format(__version__),
      )



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module that encapsulates all the platform-specific logic related to creating
named pipes."""

import os
import sys
import uuid
from threading import Thread
from typing import Optional

class Pipe:
  """A class that represents a pipe."""

  def __init__(self) -> None:
    """Initializes a non-functioning pipe."""
    
    self._read_pipe_name = ''
    self._write_pipe_name = ''
    self._thread: Optional[Thread] = None

  @staticmethod
  def create_ipc_pipe(temp_dir: str, suffix: str = '') -> 'Pipe':
    """A static method used to create a pipe between two processes. 
    
    On POSIX systems, it creates a named pipe using `os.mkfifo`.
    
    On Windows platforms, it starts a backgroud thread that transfars data from the
    writer to the reader process it is connected to.
    """

    unique_name = str(uuid.uuid4()) + suffix
    pipe = Pipe()

    if sys.platform == 'win32':
      import win32pipe # type: ignore
      pipe_name = '-nt-shaka-' + unique_name
      # The read pipe is connected to a writer process.
      pipe._read_pipe_name = r'\\.\pipe\W' + pipe_name
      # The write pipe is connected to a reader process.
      pipe._write_pipe_name = r'\\.\pipe\R' + pipe_name
      buf_size = 64 * 1024

      read_side = win32pipe.CreateNamedPipe(
          pipe._read_pipe_name,
          win32pipe.PIPE_ACCESS_INBOUND,
          win32pipe.PIPE_WAIT | win32pipe.PIPE_TYPE_BYTE | win32pipe.PIPE_READMODE_BYTE,
          1,
          buf_size,
          buf_size,
          0,
          None)

      write_side = win32pipe.CreateNamedPipe(
          pipe._write_pipe_name,
          win32pipe.PIPE_ACCESS_OUTBOUND,
          win32pipe.PIPE_WAIT | win32pipe.PIPE_TYPE_BYTE | win32pipe.PIPE_READMODE_BYTE,
          1,
          buf_size,
          buf_size,
          0,
          None)

      pipe._thread = Thread(
        target=Pipe._win_thread_fn,
        args=(read_side, write_side, buf_size),
        daemon=True)
      # Start the thread.
      pipe._thread.start()
    elif hasattr(os, 'mkfifo'):
      pipe_name = os.path.join(temp_dir, unique_name)
      pipe._read_pipe_name = pipe_name
      pipe._write_pipe_name = pipe_name
      readable_by_owner_only = 0o600  # Unix permission bits
      os.mkfifo(pipe_name, mode=readable_by_owner_only)
    else:
      raise RuntimeError('Platform not supported.')
    return pipe

  @staticmethod
  def create_file_pipe(path: str, mode: str) -> 'Pipe':
    """Returns a Pipe object whose read or write end is a path to a file."""

    pipe = Pipe()
    # A process will write on the read pipe(file).
    if mode == 'w':
      pipe._read_pipe_name = path
    # A process will read from the write pipe(file).
    elif mode == 'r':
      pipe._write_pipe_name = path
    else:
      raise RuntimeError("'{}' is not a valid mode for a Pipe.".format(mode))
    return pipe

  @staticmethod
  def _win_thread_fn(read_side, write_side, buf_size):
    """This method serves as a server that connects a writer client
    to a reader client.
    
    This methods will run as a thread, and will only be called on Windows platforms.
    """

    import win32pipe, win32file, pywintypes # type: ignore
    try:
      # Connect to both ends of the pipe before starting the transfer.
      # This funciton is blocking. If no process is connected yet, it will wait
      # indefinitely.
      win32pipe.ConnectNamedPipe(read_side)
      win32pipe.ConnectNamedPipe(write_side)
      while True:
        # Writer -> read_side -> write_side -> Reader
        _, data = win32file.ReadFile(read_side, buf_size)
        win32file.WriteFile(write_side, data)
    except Exception as ex:
      # Remove the pipes from the system.
      win32file.CloseHandle(read_side)
      win32file.CloseHandle(write_side)
      # If the error was due to one of the processes shutting down, just exit normally.
      if isinstance(ex, pywintypes.error) and ex.args[0] in [109, 232]:
        return 0
      # Otherwise, raise that error.
      raise ex

  def read_end(self) -> str:
    """Returns a pipe/file path that a reader process can read from."""
    assert self._write_pipe_name
    return self._write_pipe_name

  def write_end(self) -> str:
    """Returns a pipe/file path that a writer process can write to."""
    assert self._read_pipe_name
    return self._read_pipe_name
